from numba import njit, prange, guvectorize, float64
import numpy as np
from numpy import ndarray

__cache = True


@njit(nogil=True, cache=__cache)
def monoms_H8_single(x: ndarray) -> ndarray:
    r, s, t = x
    return np.array([1, r, s, t, r * s, r * t, s * t, r * s * t], dtype=x.dtype)


@njit(nogil=True, parallel=True, cache=__cache)
def monoms_H8_multi(x: ndarray) -> ndarray:
    nP = x.shape[0]
    res = np.zeros((nP, 8), dtype=x.dtype)
    for i in prange(nP):
        res[i] = monoms_H8_single(x[i])
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def monoms_H8_bulk_multi(x: ndarray) -> ndarray:
    nE = x.shape[0]
    res = np.zeros((nE, 8, 8), dtype=x.dtype)
    for i in prange(nE):
        res[i] = monoms_H8_multi(x[i])
    return res


def monoms_H8(x: ndarray) -> ndarray:
    N = len(x.shape)
    if N == 1:
        return monoms_H8_single(x)
    elif N == 2:
        return monoms_H8_multi(x)
    elif N == 3:
        return monoms_H8_bulk_multi(x)
    else:
        raise NotImplementedError


@njit(nogil=True, parallel=True, cache=__cache)
def _pip_H8_bulk_(x: ndarray, tol: float = 1e-12) -> ndarray:
    nE, nP = x.shape[:2]
    res = np.zeros((nP, nE), dtype=np.bool_)
    for i in prange(nP):
        for j in prange(nE):
            c1 = np.all(x[j, i] > (-1 - tol))
            c2 = np.all(x[j, i] < (1 + tol))
            res[i, j] = c1 & c2
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def _pip_H8_bulk_knn_(x: ndarray, knn: ndarray, tol: float = 1e-12) -> ndarray:
    nE, nP = x.shape[:2]
    nK = knn.shape[-1]
    res = np.zeros((nP, nE), dtype=np.bool_)
    for i in prange(nP):
        for k in prange(nK):
            c1 = np.all(x[k, i] > (-1 - tol))
            c2 = np.all(x[k, i] < (1 + tol))
            res[i, k] = c1 & c2
    return res


@njit(nogil=True, cache=__cache)
def shp_H8(pcoord: ndarray) -> ndarray:
    r, s, t = pcoord
    res = np.array(
        [
            -0.125 * r * s * t
            + 0.125 * r * s
            + 0.125 * r * t
            - 0.125 * r
            + 0.125 * s * t
            - 0.125 * s
            - 0.125 * t
            + 0.125,
            0.125 * r * s * t
            - 0.125 * r * s
            - 0.125 * r * t
            + 0.125 * r
            + 0.125 * s * t
            - 0.125 * s
            - 0.125 * t
            + 0.125,
            -0.125 * r * s * t
            + 0.125 * r * s
            - 0.125 * r * t
            + 0.125 * r
            - 0.125 * s * t
            + 0.125 * s
            - 0.125 * t
            + 0.125,
            0.125 * r * s * t
            - 0.125 * r * s
            + 0.125 * r * t
            - 0.125 * r
            - 0.125 * s * t
            + 0.125 * s
            - 0.125 * t
            + 0.125,
            0.125 * r * s * t
            + 0.125 * r * s
            - 0.125 * r * t
            - 0.125 * r
            - 0.125 * s * t
            - 0.125 * s
            + 0.125 * t
            + 0.125,
            -0.125 * r * s * t
            - 0.125 * r * s
            + 0.125 * r * t
            + 0.125 * r
            - 0.125 * s * t
            - 0.125 * s
            + 0.125 * t
            + 0.125,
            0.125 * r * s * t
            + 0.125 * r * s
            + 0.125 * r * t
            + 0.125 * r
            + 0.125 * s * t
            + 0.125 * s
            + 0.125 * t
            + 0.125,
            -0.125 * r * s * t
            - 0.125 * r * s
            - 0.125 * r * t
            - 0.125 * r
            + 0.125 * s * t
            + 0.125 * s
            + 0.125 * t
            + 0.125,
        ]
    )
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def shp_H8_multi(pcoords: ndarray) -> ndarray:
    nP = pcoords.shape[0]
    res = np.zeros((nP, 8), dtype=pcoords.dtype)
    for iP in prange(nP):
        res[iP, :] = shp_H8(pcoords[iP])
    return res


@njit(nogil=True, parallel=False, cache=__cache)
def shape_function_matrix_H8(pcoord: ndarray, ndof: int = 3) -> ndarray:
    eye = np.eye(3, dtype=pcoord.dtype)
    shp = shp_H8(pcoord)
    res = np.zeros((ndof, ndof * 8), dtype=pcoord.dtype)
    for i in prange(8):
        res[:, i * ndof : (i + 1) * ndof] = eye * shp[i]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def shape_function_matrix_H8_multi(pcoords: ndarray, ndof: int = 3) -> ndarray:
    nP = pcoords.shape[0]
    res = np.zeros((nP, ndof, ndof * 8), dtype=pcoords.dtype)
    for iP in prange(nP):
        res[iP] = shape_function_matrix_H8(pcoords[iP], ndof)
    return res


@njit(nogil=True, cache=__cache)
def dshp_H8(pcoord: ndarray) -> ndarray:
    r, s, t = pcoord
    res = np.array(
        [
            [
                -0.125 * s * t + 0.125 * s + 0.125 * t - 0.125,
                -0.125 * r * t + 0.125 * r + 0.125 * t - 0.125,
                -0.125 * r * s + 0.125 * r + 0.125 * s - 0.125,
            ],
            [
                0.125 * s * t - 0.125 * s - 0.125 * t + 0.125,
                0.125 * r * t - 0.125 * r + 0.125 * t - 0.125,
                0.125 * r * s - 0.125 * r + 0.125 * s - 0.125,
            ],
            [
                -0.125 * s * t + 0.125 * s - 0.125 * t + 0.125,
                -0.125 * r * t + 0.125 * r - 0.125 * t + 0.125,
                -0.125 * r * s - 0.125 * r - 0.125 * s - 0.125,
            ],
            [
                0.125 * s * t - 0.125 * s + 0.125 * t - 0.125,
                0.125 * r * t - 0.125 * r - 0.125 * t + 0.125,
                0.125 * r * s + 0.125 * r - 0.125 * s - 0.125,
            ],
            [
                0.125 * s * t + 0.125 * s - 0.125 * t - 0.125,
                0.125 * r * t + 0.125 * r - 0.125 * t - 0.125,
                0.125 * r * s - 0.125 * r - 0.125 * s + 0.125,
            ],
            [
                -0.125 * s * t - 0.125 * s + 0.125 * t + 0.125,
                -0.125 * r * t - 0.125 * r - 0.125 * t - 0.125,
                -0.125 * r * s + 0.125 * r - 0.125 * s + 0.125,
            ],
            [
                0.125 * s * t + 0.125 * s + 0.125 * t + 0.125,
                0.125 * r * t + 0.125 * r + 0.125 * t + 0.125,
                0.125 * r * s + 0.125 * r + 0.125 * s + 0.125,
            ],
            [
                -0.125 * s * t - 0.125 * s - 0.125 * t - 0.125,
                -0.125 * r * t - 0.125 * r + 0.125 * t + 0.125,
                -0.125 * r * s - 0.125 * r + 0.125 * s + 0.125,
            ],
        ]
    )
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def dshp_H8_multi(pcoords: ndarray) -> ndarray:
    nP = pcoords.shape[0]
    res = np.zeros((nP, 8, 3), dtype=pcoords.dtype)
    for iP in prange(nP):
        res[iP] = dshp_H8(pcoords[iP])
    return res


@njit(nogil=True, parallel=True, fastmath=True, cache=__cache)
def volumes_H8(ecoords: ndarray, qpos: ndarray, qweight: ndarray) -> ndarray:
    nE = ecoords.shape[0]
    volumes = np.zeros(nE, dtype=ecoords.dtype)
    nQ = len(qweight)
    for iQ in range(nQ):
        dshp = dshp_H8(qpos[iQ])
        for i in prange(nE):
            jac = ecoords[i].T @ dshp
            djac = np.linalg.det(jac)
            volumes[i] += qweight[iQ] * djac
    return volumes



from numba import njit, prange
from numpy import ndarray

__cache = True


@njit(nogil=True, cache=__cache)
def _approximate_multi(N: ndarray, values_source: ndarray, out: ndarray) -> ndarray:
    # N: (nP_T x nP_S)
    # values_source: (nX, nP_S)
    # out: (nX, nP_T)
    nX = out.shape[0]
    for i in prange(nX):
        out[i] = N @ values_source[i]



# -*- coding: utf-8 -*-
from functools import partial

from ..geometry import PolyCellGeometry1d
from ..data.polycell import PolyCell
from ..utils.numint import Gauss_Legendre_Line_Grid
from ..utils.cells.l3 import monoms_L3


__all__ = ["L3"]


class L3(PolyCell):
    """
    3-Node line element.
    """

    class Geometry(PolyCellGeometry1d):
        number_of_nodes = 3
        vtk_cell_id = 21
        monomial_evaluator = monoms_L3
        quadrature = {
            "full": partial(Gauss_Legendre_Line_Grid, 3),
            "geometry": "full",
        }



# -*- coding: utf-8 -*-
from typing import Tuple, Optional, Union, Iterable
from numbers import Number

import numpy as np
from numpy import ndarray

from sigmaepsilon.math.utils import atleast2d

from .data import PolyData, PointData
from .space import CartesianFrame
from .cells import H8
from .grid import grid
from .utils.topology import detach_mesh_bulk
from .utils import cell_centers_bulk, cells_coords
from .utils.knn import k_nearest_neighbours
from .utils.cic import H8_in_TET4_bulk_knn


__all__ = ["voxelize_cylinder", "voxelize_TET4_H8"]


def voxelize_cylinder(
    radius: ndarray | Number | Iterable[Number],
    height: float,
    size: float,  # voxel edge length
    frame: Optional[Union[CartesianFrame, None]] = None,
) -> Tuple[ndarray, ndarray]:
    """
    Returns raw mesh data of a voxelized cylinder.

    Parameters
    ----------
    radius: numpy.ndarray or Number or Iterable[Number]
        Radius of the cylinder. If a NumPy array is passed, it should
        contain two values, the inner and outer radii. If a single
        number is passed, it is assumed to be the outer radius.
    height: float
        Height of the cylinder.
    size: float
        Size of the voxel grid.
    frame: CartesianFrame or None, Optional
        Cartesian frame of the voxel grid. Default is None.

    Example
    -------
    The following example shows how to create a voxelized cylinder.
    >>> import numpy as np
    >>> from sigmaepsilon.mesh.voxelize import voxelize_cylinder
    >>> radius = 1
    >>> height = 2
    >>> size = 0.1
    >>> coords, topo = voxelize_cylinder(radius, height, size)
    """
    if isinstance(radius, int):
        radius = np.array([0, radius])
    elif not isinstance(radius, ndarray):
        radius = np.array(radius)

    nXY = int(np.ceil(2 * radius[1] / size))
    nZ = int(np.ceil(height / size))
    Lxy, Lz = 2 * radius[1], height
    coords, topo = grid(
        size=(Lxy, Lxy, Lz), shape=(nXY, nXY, nZ), eshape="H8", centralize=True
    )
    frame = CartesianFrame(dim=3) if frame is None else frame
    pd = PointData(coords=coords, frame=frame)
    cd = H8(topo=topo, frames=frame)
    c = PolyData(pd, cd).centers()
    r = (c[:, 0] ** 2 + c[:, 1] ** 2) ** (1 / 2)
    cond = (r <= radius[1]) & (r >= radius[0])
    inds = np.where(cond)[0]
    return detach_mesh_bulk(coords, topo[inds])


def voxelize_TET4_H8(
    coords_TET4: ndarray,
    topo_TET4: ndarray,
    shape: tuple | None = None,
    resolution: float | None = None,
    k_max: int = 4,
    tol: float = 1e-12,
) -> Tuple[ndarray, ndarray]:
    """
    Returns a voxelized version of a tetrahadral mesh.

    The function is expected to behave well, if the input mesh is
    regular. If it contains extremely skew cells, the function may
    struggle to find the correct voxelization.

    Parameters
    ----------
    coords_TET4: numpy.ndarray
        2d NumPy array of the coordinates of the nodes of the TET4 cells.
    topo_TET4: numpy.ndarray
        2d NumPy array of the topology of the TET4 cells.
    shape: tuple, Optional
        Tuple of the shape of the voxel grid. Default is None.
    resolution: float, Optional
        Resolution of the voxel grid. Default is None.
    k_max: int, Optional
        Maximum number of nearest neighbours to consider. Default is 4.
        If the number of TET4 cells is less than `k_max`, the function
        will use the number of TET4 cells as the `k` parameter.
    tol: float, Optional
        Tolerance to consider a point inside a cell. Default is 1e-12.

    Example
    -------
    The following example shows how to voxelize a TET4 mesh.

    >>> import numpy as np
    >>> from sigmaepsilon.mesh.voxelize import voxelize_TET4_H8
    >>> coords_TET4 = np.array([[0, 0, 0], [1, 0, 0], [0, 1, 0], [0, 0, 1]]).astype(float)
    >>> topo_TET4 = np.array([[0, 1, 2, 3]])
    >>> shape = (10, 10, 10)
    >>> coords_H8, topo_H8 = voxelize_TET4_H8(coords_TET4, topo_TET4, shape=shape)
    """
    size_x = np.max(coords_TET4[:, 0]) - np.min(coords_TET4[:, 0])
    size_y = np.max(coords_TET4[:, 1]) - np.min(coords_TET4[:, 1])
    size_z = np.max(coords_TET4[:, 2]) - np.min(coords_TET4[:, 2])
    size = (size_x, size_y, size_z)

    shift_x = np.min(coords_TET4[:, 0])
    shift_y = np.min(coords_TET4[:, 1])
    shift_z = np.min(coords_TET4[:, 2])
    shift = (shift_x, shift_y, shift_z)

    if shape is None and resolution is not None:
        n_x = int(np.ceil(size_x / resolution))
        n_y = int(np.ceil(size_y / resolution))
        n_z = int(np.ceil(size_z / resolution))
        shape = (n_x, n_y, n_z)

    coords_H8, topo_H8 = grid(size=size, shape=shape, eshape="H8", shift=shift)
    centers_H8 = cell_centers_bulk(coords_H8, topo_H8)
    centers_TET4 = cell_centers_bulk(coords_TET4, topo_TET4)
    k = min(k_max, len(centers_TET4))
    neighbours = k_nearest_neighbours(centers_TET4, centers_H8, k=k)
    neighbours = atleast2d(neighbours, back=True)
    cell_coords_TET4 = cells_coords(coords_TET4, topo_TET4)
    cell_coords_H8 = cells_coords(coords_H8, topo_H8)
    H8_bool = H8_in_TET4_bulk_knn(cell_coords_H8, cell_coords_TET4, neighbours, tol=tol)
    return coords_H8, topo_H8[H8_bool]



from typing import Iterable
import numpy as np

from awkward import Array as akArray

from sigmaepsilon.math.linalg.sparse import JaggedArray
from sigmaepsilon.math.arraysetops import unique2d
from sigmaepsilon.math import atleast2d


__all__ = ["TopologyArray"]


HANDLED_FUNCTIONS = {}


class TopologyArray(JaggedArray):
    """
    A class to handle complex topologies. It is a subclass of
    the JaggedArray class from the Neumann library and is compatible
    with Numpy universal functions.

    Parameters
    ----------
    *topo: Iterable
        One or more 2d arrays definig topologies for polygonal cells.
    cuts: Iterable, Optional
        An iterable that tells how to unflatten an 1d array into a
        2d jagged shape. Only if topology is provided as a 1d array.
        Default is None.
    force_numpy: bool, Optional
        Forces dense inputs to be NumPy arrays in the background.
        Default is True.

    Examples
    --------
    The following could be the definiton for a mesh consisting from
    two line cells, one with 3 nodes and another with 2:

    >>> import numpy as np
    >>> from sigmaepsilon.mesh import TopologyArray
    >>> data = np.array([0, 1, 2, 3, 4])
    >>> TopologyArray(data, cuts=[3, 2])
    TopologyArray([[0, 1, 2], [3, 4]])

    The same mesh defined in another way:

    >>> topo1 = np.array([0, 1, 2])
    >>> topo2 = np.array([3, 4])
    >>> TopologyArray(topo1, topo2)
    TopologyArray([[0, 1, 2], [3, 4]])

    Let assume we have two 4-noded quadrilaterals as well:

    >>> topo3 = np.array([[5, 6, 7, 8],[6, 7, 9, 10]])
    >>> TopologyArray(topo1, topo2, topo3)
    TopologyArray([[0, 1, 2], [3, 4], [5, 6, 7, 8], [6, 7, 9, 10]])

    Since the TopologyArray class is a subclass of JaggedArray, we
    can easily transform it to a CSR matrix, or an Awkward array:

    >>> TopologyArray(topo1, topo2, topo3).to_csr()
    4x4 CSR matrix of 13 values.

    >>> ak_array = TopologyArray(topo1, topo2, topo3).to_ak()

    To get the unique indices in a mesh, you can simply use NumPy:

    >>> t = TopologyArray(topo1, topo2, topo3)
    >>> np.unique(t)
    array([ 0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10])

    You can also combine different topologies into one object by
    stacking:

    >>> t1 = TopologyArray(topo1, topo2)
    >>> t2 = TopologyArray(topo3)
    >>> np.vstack([t1, t2])
    TopologyArray([[0, 1, 2], [3, 4], [5, 6, 7, 8], [6, 7, 9, 10]])

    See Also
    --------
    :class:`~sigmaepsilon.math.linalg.sparse.jaggedarray.JaggedArray`
    """

    def __init__(self, *topo, cuts: Iterable = None, force_numpy: bool = True):
        if len(topo) == 1 and cuts is None:
            if isinstance(topo[0], np.ndarray):
                data = atleast2d(topo[0], front=True)
            elif isinstance(topo[0], akArray):
                data = topo[0]
        elif len(topo) == 1 and cuts is not None:
            data = np.array(topo[0]).astype(int)
            cuts = np.array(cuts).astype(int)
        else:
            topo = list(map(lambda t: atleast2d(t, front=True), topo))
            widths = list(map(lambda topo: topo.shape[1], topo))
            widths = np.array(widths, dtype=int)
            cN, cE = 0, 0
            for i in range(len(topo)):
                dE = topo[i].shape[0]
                cE += dE
                cN += dE * topo[i].shape[1]
            data = np.zeros(cN, dtype=int)
            cuts = np.zeros(cE, dtype=int)
            cN, cE = 0, 0
            for i in range(len(topo)):
                dE = topo[i].shape[0]
                dN = dE * topo[i].shape[1]
                data[cN : cN + dN] = topo[i].flatten()
                cN += dN
                cuts[cE : cE + dE] = np.full(dE, widths[i])
                cE += dE
        super(TopologyArray, self).__init__(data, cuts=cuts, force_numpy=force_numpy)

    def __array_function__(self, func, types, args, kwargs):
        if func not in HANDLED_FUNCTIONS:
            arrs = [arg._wrapped for arg in args]
            return func(*arrs, **kwargs)
        # Note: this allows subclasses that don't override
        # __array_function__ to handle DiagonalArray objects.
        if not all(issubclass(t, self.__class__) for t in types):
            return NotImplemented
        return HANDLED_FUNCTIONS[func](*args, **kwargs)


def implements(numpy_function):
    """
    Register an __array_function__ implementation for
    TopologyArray objects.
    """

    def decorator(func):
        HANDLED_FUNCTIONS[numpy_function] = func
        return func

    return decorator


@implements(np.unique)
def unique(*args, **kwargs):
    return unique2d(args[0]._wrapped, **kwargs)


@implements(np.vstack)
def vstack(*args, **kwargs):
    data = np.concatenate(list(t.flatten() for t in args[0]))
    cuts = np.concatenate(list(t.widths() for t in args[0]))
    return TopologyArray(data, cuts=cuts)



from numba import njit, prange
import numpy as np
from numpy import ndarray
from sigmaepsilon.math import flatten2dC

from ..tri import area_tri_bulk


__cache = True


@njit(nogil=True, cache=__cache)
def monoms_Q4(x: ndarray) -> ndarray:
    r, s = x
    return np.array([1, r, s, r * s], dtype=float)


@njit(nogil=True, parallel=True, cache=__cache)
def area_Q4_bulk(ecoords: np.ndarray):
    nE = len(ecoords)
    res = np.zeros(nE, dtype=ecoords.dtype)
    res += area_tri_bulk(ecoords[:, :3, :])
    res += area_tri_bulk(ecoords[:, np.array([0, 2, 3]), :])
    return res


@njit(nogil=True, cache=__cache)
def shp_Q4(pcoord: np.ndarray):
    r, s = pcoord[:2]
    res = np.array(
        [
            [0.25 * (1 - r) * (1 - s)],
            [0.25 * (1 + r) * (1 - s)],
            [0.25 * (1 + r) * (1 + s)],
            [0.25 * (1 - r) * (1 + s)],
        ],
        dtype=pcoord.dtype,
    )
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def shp_Q4_multi(pcoords: np.ndarray):
    nP = pcoords.shape[0]
    res = np.zeros((nP, 4), dtype=pcoords.dtype)
    for iP in prange(nP):
        res[iP, :] = flatten2dC(shp_Q4(pcoords[iP]))
    return res


@njit(nogil=True, cache=__cache)
def dshp_Q4(pcoord: ndarray):
    r, s = pcoord[:2]
    res = np.array(
        [
            [(s - 1) / 4, (r - 1) / 4],
            [(1 - s) / 4, (-r - 1) / 4],
            [(s + 1) / 4, (r + 1) / 4],
            [(-s - 1) / 4, (1 - r) / 4],
        ],
        dtype=pcoord.dtype,
    )
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def dshp_Q4_multi(pcoords: ndarray):
    nP = pcoords.shape[0]
    res = np.zeros((nP, 4, 2), dtype=pcoords.dtype)
    for iP in prange(nP):
        res[iP] = dshp_Q4(pcoords[iP])
    return res


@njit(nogil=True, parallel=False, cache=__cache)
def shape_function_matrix_Q4(pcoord: ndarray, ndof: int = 2) -> ndarray:
    eye = np.eye(ndof, dtype=pcoord.dtype)
    shp = shp_Q4(pcoord)
    res = np.zeros((ndof, ndof * 4), dtype=pcoord.dtype)
    for i in prange(4):
        res[:, i * ndof : (i + 1) * ndof] = eye * shp[i]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def shape_function_matrix_Q4_multi(pcoords: ndarray, ndof: int = 2) -> ndarray:
    nP = pcoords.shape[0]
    res = np.zeros((nP, ndof, ndof * 4), dtype=pcoords.dtype)
    for iP in prange(nP):
        res[iP] = shape_function_matrix_Q4(pcoords[iP], ndof)
    return res



from numba import njit, prange
import numpy as np
from numpy import ndarray
from sigmaepsilon.math import flatten2dC

__cache = True


@njit(nogil=True, cache=__cache)
def monoms_Q9(x: ndarray) -> ndarray:
    r, s = x
    res = np.array(
        [
            1,
            r,
            s,
            r * s,
            r**2,
            s**2,
            r * s**2,
            s * r**2,
            s**2 * r**2,
        ],
        dtype=float,
    )
    return res


@njit(nogil=True, cache=__cache)
def shp_Q9(pcoord: np.ndarray):
    r, s = pcoord[:2]
    res = np.array(
        [
            [0.25 * r**2 * s**2 - 0.25 * r**2 * s - 0.25 * r * s**2 + 0.25 * r * s],
            [0.25 * r**2 * s**2 - 0.25 * r**2 * s + 0.25 * r * s**2 - 0.25 * r * s],
            [0.25 * r**2 * s**2 + 0.25 * r**2 * s + 0.25 * r * s**2 + 0.25 * r * s],
            [0.25 * r**2 * s**2 + 0.25 * r**2 * s - 0.25 * r * s**2 - 0.25 * r * s],
            [-0.5 * r**2 * s**2 + 0.5 * r**2 * s + 0.5 * s**2 - 0.5 * s],
            [-0.5 * r**2 * s**2 + 0.5 * r**2 - 0.5 * r * s**2 + 0.5 * r],
            [-0.5 * r**2 * s**2 - 0.5 * r**2 * s + 0.5 * s**2 + 0.5 * s],
            [-0.5 * r**2 * s**2 + 0.5 * r**2 + 0.5 * r * s**2 - 0.5 * r],
            [1.0 * r**2 * s**2 - 1.0 * r**2 - 1.0 * s**2 + 1.0],
        ],
        dtype=pcoord.dtype,
    )
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def shp_Q9_multi(pcoords: np.ndarray):
    nP = pcoords.shape[0]
    res = np.zeros((nP, 9), dtype=pcoords.dtype)
    for iP in prange(nP):
        res[iP, :] = flatten2dC(shp_Q9(pcoords[iP]))
    return res


@njit(nogil=True, parallel=False, cache=__cache)
def shape_function_matrix_Q9(pcoord: np.ndarray, ndof: int = 2):
    eye = np.eye(ndof, dtype=pcoord.dtype)
    shp = shp_Q9(pcoord)
    res = np.zeros((ndof, ndof * 9), dtype=pcoord.dtype)
    for i in range(9):
        res[:, i * ndof : (i + 1) * ndof] = eye * shp[i]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def shape_function_matrix_Q9_multi(pcoords: np.ndarray, ndof: int = 2):
    nP = pcoords.shape[0]
    res = np.zeros((nP, ndof, ndof * 9), dtype=pcoords.dtype)
    for iP in prange(nP):
        res[iP] = shape_function_matrix_Q9(pcoords[iP], ndof)
    return res


@njit(nogil=True, cache=__cache)
def dshp_Q9(pcoord: np.ndarray):
    r, s = pcoord[:2]
    res = np.array(
        [
            [
                0.5 * r * s**2 - 0.5 * r * s - 0.25 * s**2 + 0.25 * s,
                0.5 * r**2 * s - 0.25 * r**2 - 0.5 * r * s + 0.25 * r,
            ],
            [
                0.5 * r * s**2 - 0.5 * r * s + 0.25 * s**2 - 0.25 * s,
                0.5 * r**2 * s - 0.25 * r**2 + 0.5 * r * s - 0.25 * r,
            ],
            [
                0.5 * r * s**2 + 0.5 * r * s + 0.25 * s**2 + 0.25 * s,
                0.5 * r**2 * s + 0.25 * r**2 + 0.5 * r * s + 0.25 * r,
            ],
            [
                0.5 * r * s**2 + 0.5 * r * s - 0.25 * s**2 - 0.25 * s,
                0.5 * r**2 * s + 0.25 * r**2 - 0.5 * r * s - 0.25 * r,
            ],
            [
                -1.0 * r * s**2 + 1.0 * r * s,
                -1.0 * r**2 * s + 0.5 * r**2 + 1.0 * s - 0.5,
            ],
            [
                -1.0 * r * s**2 + 1.0 * r - 0.5 * s**2 + 0.5,
                -1.0 * r**2 * s - 1.0 * r * s,
            ],
            [
                -1.0 * r * s**2 - 1.0 * r * s,
                -1.0 * r**2 * s - 0.5 * r**2 + 1.0 * s + 0.5,
            ],
            [
                -1.0 * r * s**2 + 1.0 * r + 0.5 * s**2 - 0.5,
                -1.0 * r**2 * s + 1.0 * r * s,
            ],
            [2.0 * r * s**2 - 2.0 * r, 2.0 * r**2 * s - 2.0 * s],
        ],
        dtype=pcoord.dtype,
    )
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def dshp_Q9_multi(pcoords: ndarray):
    """
    Returns the first orderderivatives of the shape functions,
    evaluated at multiple points, according to 'pcoords'.

    ---
    (nP, nNE, 2)
    """
    nP = pcoords.shape[0]
    res = np.zeros((nP, 9, 2), dtype=pcoords.dtype)
    for iP in prange(nP):
        res[iP] = dshp_Q9(pcoords[iP])
    return res


if __name__ == "__main__":
    import numpy as np

    shape_function_matrix_Q9(np.array([0, 0]))
    shape_function_matrix_Q9(np.array([0, 0]))
    shape_function_matrix_Q9(np.array([0, 0]))



from typing import Tuple, List

import numpy as np
from numpy import ndarray
from sympy import symbols

from ..geometry import PolyCellGeometry3d
from ..data.polycell import PolyCell
from ..utils.numint import Gauss_Legendre_Wedge_3x2
from ..utils.cells.w6 import monoms_W6


class W6(PolyCell):
    """
    Polyhedra class for 6-noded trilinear wedges.
    """

    label = "W6"

    class Geometry(PolyCellGeometry3d):
        number_of_nodes = 6
        vtk_cell_id = 13
        monomial_evaluator = monoms_W6
        quadrature = {
            "full": Gauss_Legendre_Wedge_3x2,
            "geometry": "full",
        }

        @classmethod
        def polybase(cls) -> Tuple[List]:
            """
            Retruns the polynomial base of the master element.

            Returns
            -------
            list
                A list of SymPy symbols.
            list
                A list of monomials.
            """
            locvars = r, s, t = symbols("r s t", real=True)
            monoms = [1, r, s, t, r * t, s * t]
            return locvars, monoms

        @classmethod
        def master_coordinates(cls) -> ndarray:
            return np.array(
                [
                    [-1 / 3, -1 / 3, -1.0],
                    [2 / 3, -1 / 3, -1.0],
                    [-1 / 3, 2 / 3, -1.0],
                    [-1 / 3, -1 / 3, 1.0],
                    [2 / 3, -1 / 3, 1.0],
                    [-1 / 3, 2 / 3, 1.0],
                ]
            )

        @classmethod
        def master_center(cls) -> ndarray:
            return np.array([[0.0, 0.0, 0.0]], dtype=float)

        @classmethod
        def tetmap(cls) -> ndarray:
            return np.array(
                [[0, 1, 2, 4], [3, 5, 4, 2], [2, 5, 0, 4]],
                dtype=int,
            )



# -*- coding: utf-8 -*-
from ...config import __hasmatplotlib__

if not __hasmatplotlib__:  # pragma: no cover

    def triplotter(*_, **__):
        raise ImportError(
            "You need Matplotlib for this. Install it with 'pip install matplotlib'. "
            "You may also need to restart your kernel and reload the package."
        )

    def get_fig_axes(*_, **__):
        raise ImportError(
            "You need Matplotlib for this. Install it with 'pip install matplotlib'. "
            "You may also need to restart your kernel and reload the package."
        )

    def decorate_mpl_ax(*_, **__):
        raise ImportError(
            "You need Matplotlib for this. Install it with 'pip install matplotlib'. "
            "You may also need to restart your kernel and reload the package."
        )

else:
    from typing import Iterable, Callable, Any
    from functools import wraps

    import numpy as np
    from numpy import ndarray
    import matplotlib.pyplot as plt
    from matplotlib.figure import Figure, Axes
    from matplotlib.patches import Polygon
    from matplotlib.collections import PatchCollection

    from sigmaepsilon.mesh.triang import triangulate
    from sigmaepsilon.core.typing import issequence

    class TriPatchCollection(PatchCollection):
        def __init__(self, cellcoords, *args, **kwargs):
            pmap = map(lambda i: cellcoords[i], np.arange(len(cellcoords)))

            def fnc(points):
                return Polygon(points, closed=True)

            patches = list(map(fnc, pmap))
            super().__init__(patches, *args, **kwargs)

    def triplotter(plotter: Callable) -> Callable:
        @wraps(plotter)
        def inner(
            triobj: Any,
            *args,
            data: ndarray = None,
            title: str = None,
            label: str = None,
            fig: Figure = None,
            ax: Axes = None,
            axes: Iterable[Axes] = None,
            fig_kw: dict = None,
            **kwargs,
        ) -> Any:
            fig, axes = get_fig_axes(
                *args, data=data, ax=ax, axes=axes, fig=fig, fig_kw=fig_kw
            )

            if isinstance(triobj, tuple):
                coords, topo = triobj
                triobj = triangulate(points=coords[:, :2], triangles=topo)[-1]
                coords, topo = None, None

            if data is not None:
                if not len(data.shape) <= 2:
                    raise ValueError("Data must be a 1 or 2 dimensional array.")

                nD = 1 if len(data.shape) == 1 else data.shape[1]

                data = data.reshape((data.shape[0], nD))

                if not issequence(title):
                    title = nD * (title,)

                if not issequence(label):
                    label = nD * (label,)

                axobj = [
                    plotter(
                        triobj,
                        data[:, i],
                        fig=fig,
                        ax=ax,
                        title=title[i],
                        label=label[i],
                        **kwargs,
                    )
                    for i, ax in enumerate(axes)
                ]
                if nD == 1:
                    data = data.reshape(data.shape[0])
            else:
                axobj = plotter(triobj, ax=axes[0], fig=fig, title=title, **kwargs)

            return axobj

        return inner

    def get_fig_axes(
        *args,
        data=None,
        fig=None,
        axes=None,
        shape=None,
        horizontal=False,
        ax=None,
        fig_kw=None,
    ) -> tuple:
        """
        Returns a figure and an axes object.
        """
        if isinstance(ax, (tuple, list)):
            axes = ax

        if fig is not None:
            if axes is not None:
                return fig, axes
            elif ax is not None:
                return fig, (ax,)
        else:
            if fig_kw is None:
                fig_kw = {}

            if data is not None:
                nD = 1 if len(data.shape) == 1 else data.shape[1]

                if nD == 1:
                    try:
                        ax = args[0]
                    except Exception:
                        fig, ax = plt.subplots(**fig_kw)
                    return fig, (ax,)

                if fig is None or axes is None:
                    if shape is not None:
                        if isinstance(shape, int):
                            shape = (shape, 1) if horizontal else (1, shape)
                        assert nD == (
                            shape[0] * shape[1]
                        ), "Mismatch in shape and data."
                    else:
                        shape = (nD, 1) if horizontal else (1, nD)

                    fig, axes = plt.subplots(*shape, **fig_kw)

                if not isinstance(axes, Iterable):
                    axes = (axes,)

                return fig, axes
            else:
                try:
                    ax = args[0]
                except Exception:
                    fig, ax = plt.subplots(**fig_kw)
                return fig, (ax,)

        return None, None

    def decorate_mpl_ax(
        *,
        fig=None,
        ax=None,
        aspect="equal",
        xlim=None,
        ylim=None,
        axis="on",
        offset=0.05,
        points=None,
        axfnc: Callable = None,
        title=None,
        suptitle=None,
        label=None,
    ):
        """
        Decorates an axis using the most often used modifiers.
        """
        assert ax is not None, (
            "A matplotlib Axes object must be provided with " "keyword argument 'ax'!"
        )

        if axfnc is not None:
            try:
                axfnc(ax)
            except Exception:
                raise RuntimeError("Something went wrong when calling axfnc.")

        if xlim is None:
            if points is not None:
                xlim = points[:, 0].min(), points[:, 0].max()
                if offset is not None:
                    dx = np.abs(xlim[1] - xlim[0])
                    xlim = xlim[0] - offset * dx, xlim[1] + offset * dx

        if ylim is None:
            if points is not None:
                ylim = points[:, 1].min(), points[:, 1].max()
                if offset is not None:
                    dx = np.abs(ylim[1] - ylim[0])
                    ylim = ylim[0] - offset * dx, ylim[1] + offset * dx

        ax.set_aspect(aspect)
        ax.axis(axis)

        if xlim is not None:
            ax.set_xlim(*xlim)

        if ylim is not None:
            ax.set_ylim(*ylim)

        if title is not None:
            ax.set_title(title)

        if label is not None:
            ax.set_xlabel(label)

        if fig is not None and suptitle is not None:
            fig.suptitle(suptitle)

        return ax



from typing import Union, Iterable, List
from copy import copy

import numpy as np
from numpy import ndarray


class IndexManager:
    """
    Manages and index set by generating and recycling indices.

    Examples
    --------
    >>> from sigmaepsilon.mesh.indexmanager import IndexManager
    >>> im = IndexManager()
    >>> im.generate()
    0
    >>> im.generate(1)
    1
    >>> im.generate(2)
    [2, 3]
    >>> im.recycle(0)
    >>> im.generate(2)
    [0, 4]
    >>> im.generate_np(2)
    array([5, 6])
    >>> im.recycle(5, 6)
    >>> im.generate_np(2)
    array([5, 6])
    >>> im.recycle([5, 6])
    >>> im.generate_np(2)
    array([5, 6])
    """

    def __init__(self, start: int = 0):
        self.queue = []
        self.next = start

    def generate_np(self, n: int = 1) -> Union[int, ndarray]:
        """
        Generates indices as NumPy arrays.
        """
        if n == 1:
            return self.generate(1)
        else:
            return np.array(self.generate(n))

    def generate(self, n: int = 1) -> Union[int, List[int]]:
        """
        Generates one or more indices.
        """
        nQ = len(self.queue)
        if nQ > 0:
            if n == 1:
                res = self.queue.pop()
            else:
                if nQ >= n:
                    res = self.queue[:n]
                    del self.queue[:n]
                else:
                    res = copy(self.queue)
                    res.extend(range(self.next, self.next + n - nQ))
                    self.queue = []
                self.next += n - nQ
        else:
            if n == 1:
                res = self.next
            else:
                res = list(range(self.next, self.next + n))
            self.next += n
        return res

    def recycle(self, *args) -> None:
        """
        Recycles some indices.
        """
        for a in args:
            if isinstance(a, Iterable):
                self.queue.extend(a)
            else:
                self.queue.append(a)



from typing import Union, Tuple, Iterable

import numpy as np
from numpy import ndarray
from numpy.linalg import norm
from numba import njit, prange
from numba.typed import Dict as nbDict
from numba import types as nbtypes

from sigmaepsilon.math import matrixform, atleast2d
from sigmaepsilon.math.linalg.sparse import JaggedArray, csr_matrix

__cache = True
nbint64 = nbtypes.int64
nbint64A = nbint64[:]
nbfloat64A = nbtypes.float64[:]


def cells_around(*args, **kwargs) -> Union[JaggedArray, csr_matrix, dict]:
    """
    Alias for :func:`points_around`.
    """
    return points_around(*args, **kwargs)


def points_around(
    points: ndarray,
    r_max: float,
    *,
    frmt: str = "dict",
    MT: bool = True,
    n_max: int = 10,
) -> Union[JaggedArray, csr_matrix, dict]:
    """
    Returns neighbouring points for each entry in `points` that are
    closer than the distance `r_max`. The results are returned in
    diffent formats, depending on the format specifier argument `frmt`.

    Parameters
    ----------
    points: numpy.ndarray
        Coordinates of several points as a 2d float numpy array.
    r_max: float
        Maximum distance.
    n_max: int, Optional
        Maximum number of neighbours. Default is 10.
    frmt: str
        A string specifying the output format. Valid options are
        'jagged', 'csr' and 'dict'.
        See below for the details on the returned object.

    Returns
    -------
    if frmt = 'csr' : sigmaepsilon.math.linalg.sparse.csr.csr_matrix
        A numba-jittable sparse matrix format.

    frmt = 'dict' : numba Dict(int : int[:])

    frmt = 'jagged' : sigmaepsilon.math.linalg.sparse.JaggedArray
        A subclass of `awkward.Array`
    """
    if MT:
        data, widths = _cells_around_MT_(points, r_max, n_max)
    else:
        raise NotImplementedError

    if frmt == "dict":
        return _cells_data_to_dict(data, widths)
    elif frmt == "jagged":
        return _cells_data_to_jagged(data, widths)
    elif frmt == "csr":
        d = _cells_data_to_dict(data, widths)
        data, inds, indptr, shp = _dict_to_spdata(d, widths)
        return csr_matrix(data=data, indices=inds, indptr=indptr, shape=shp)

    raise RuntimeError("Unhandled case!")


@njit(nogil=True, cache=__cache)
def _dict_to_spdata(d: dict, widths: np.ndarray):
    N = int(np.sum(widths))
    nE = len(widths)
    data = np.zeros(N, dtype=np.int64)
    inds = np.zeros_like(data)
    indptr = np.zeros(nE + 1, dtype=np.int64)
    _c = 0
    wmax = 0
    for i in range(len(d)):
        w = widths[i]
        if w > wmax:
            wmax = w
        c_ = _c + w
        data[_c:c_] = d[i]
        inds[_c:c_] = np.arange(w)
        indptr[i + 1] = c_
        _c = c_
    return data, inds, indptr, (nE, wmax)


@njit(nogil=True, cache=__cache)
def _jagged_to_spdata(ja: JaggedArray):
    widths = ja.widths()
    N = int(np.sum(widths))
    nE = len(widths)
    data = np.zeros(N, dtype=np.int64)
    inds = np.zeros_like(data)
    indptr = np.zeros(nE + 1, dtype=np.int64)
    _c = 0
    wmax = 0
    for i in range(len(ja)):
        w = widths[i]
        if w > wmax:
            wmax = w
        c_ = _c + w
        data[_c:c_] = ja[i]
        inds[_c:c_] = np.arange(w)
        indptr[i + 1] = c_
        _c = c_
    return data, inds, indptr, (nE, wmax)


@njit(nogil=True, fastmath=True, cache=__cache)
def _cells_data_to_dict(data: np.ndarray, widths: np.ndarray) -> nbDict:
    dres = dict()
    nE = len(widths)
    for iE in range(nE):
        dres[iE] = data[iE, : widths[iE]]
    return dres


@njit(nogil=True, parallel=True, cache=__cache)
def _flatten_jagged_data(data, widths) -> ndarray:
    nE = len(widths)
    inds = np.zeros(nE + 1, dtype=widths.dtype)
    inds[1:] = np.cumsum(widths)
    res = np.zeros(np.sum(widths))
    for i in prange(nE):
        res[inds[i] : inds[i + 1]] = data[i, : widths[i]]
    return res


def _cells_data_to_jagged(data, widths):
    data = _flatten_jagged_data(data, widths)
    return JaggedArray(data, cuts=widths)


@njit(nogil=True, parallel=True, cache=__cache)
def _cells_around_MT_(
    centers: ndarray, r_max: float, n_max: int = 10
) -> Tuple[ndarray, ndarray]:
    nE = len(centers)
    res = np.zeros((nE, n_max), dtype=np.int64)
    widths = np.zeros(nE, dtype=np.int64)
    for iE in prange(nE):
        inds = np.where(norms(centers - centers[iE]) <= r_max)[0]
        if inds.shape[0] <= n_max:
            res[iE, : inds.shape[0]] = inds
            widths[iE] = len(inds)
        else:
            res[iE, :] = inds[:n_max]
            widths[iE] = n_max
    return res, widths


def points_of_cells(
    coords: ndarray,
    topo: ndarray,
    *,
    local_axes: ndarray = None,
    centralize: bool = True,
    centers: ndarray = None,
) -> ndarray:
    """
    Returns an explicit representation of coordinates of the cells from a
    pointset and a topology. If coordinate frames are provided, the
    coorindates are returned with  respect to those frames.

    Parameters
    ----------
    coords: numpy.ndarray
        2d float array of shape (nP, nD) of vertex coordinates.
        nP : number of points
        nD : number of dimensions of the model space
    topo: numpy.ndarray
        A 2D array of shape (nE, nNE) of vertex indices. The i-th row
        contains the vertex indices of the i-th element.
        nE : number of elements
        nNE : number of nodes per element
    local_axes: numpy.ndarray
        Reference frames as a 3d array of shape (..., 3, 3). A single
        3x3 numpy array or matrices for all elements in 'topo' must be
        provided.
    centralize: bool, Optional
        If True, and 'local_axes' is not None, the local coordinates are
        returned with respect to the geometric center of each element.
    centers: numpy.ndarray
        Centers for all cells. This is to account for different master cells
        with different centers (usually for triangles). Default is None.

    Returns
    -------
    numpy.ndarray
        3d float array of coordinates.

    Notes
    -----
    It is assumed that all entries in 'coords' are coordinates of
    points in the same frame.
    """
    if local_axes is not None:
        if centralize:
            if centers is not None:
                ec = _centralize_cells_coords_2(cells_coords(coords, topo), centers)
            else:
                ec = _centralize_cells_coords(cells_coords(coords, topo))
        else:
            ec = cells_coords(coords, topo)
        return _cells_coords_tr_(ec, local_axes)
    else:
        if centralize:
            if centers is not None:
                return _centralize_cells_coords_2(cells_coords(coords, topo), centers)
            else:
                return _centralize_cells_coords(cells_coords(coords, topo))
        else:
            return cells_coords(coords, topo)


@njit(nogil=True, parallel=True, cache=__cache)
def _cells_coords_tr_(ecoords: ndarray, local_axes: ndarray) -> ndarray:
    nE, nNE, _ = ecoords.shape
    res = np.zeros_like(ecoords)
    for i in prange(nE):
        dcm = local_axes[i]
        for j in prange(nNE):
            res[i, j, :] = dcm @ ecoords[i, j, :]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def _centralize_cells_coords(ecoords: ndarray) -> ndarray:
    nE, nNE, _ = ecoords.shape
    res = np.zeros_like(ecoords)
    for i in prange(nE):
        cc = cell_center(ecoords[i])
        for j in prange(nNE):
            res[i, j, :] = ecoords[i, j, :] - cc
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def _centralize_cells_coords_2(ecoords: ndarray, centers: ndarray) -> ndarray:
    nE, nNE, _ = ecoords.shape
    res = np.zeros_like(ecoords)
    for i in prange(nE):
        cc = centers[i]
        for j in prange(nNE):
            res[i, j, :] = ecoords[i, j, :] - cc
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def cells_coords(coords: ndarray, topo: ndarray) -> ndarray:
    """
    Returns coordinates of cells from a coordinate base array and
    a topology array.

    Parameters
    ----------
    coords: numpy.ndarray
        2d float array of shape (nP, nD) of vertex coordinates.
        nP : number of points
        nD : number of dimensions of the model space
    topo: numpy.ndarray
        A 2D array of shape (nE, nNE) of vertex indices. The i-th
        row contains the vertex indices of the i-th element.
        nE : number of elements
        nNE : number of nodes per element

    Returns
    -------
    numpy.ndarray
        A 3d array of shape (nE, nNE, nD) that contains coordinates
        for all nodes of all cells according to the argument 'topo'.

    Notes
    -----
    The array 'coords' must be fully populated up to the maximum
    index in 'topo'. (len(coords) >= (topo.max() + 1))
    """
    nE, nNE = topo.shape
    res = np.zeros((nE, nNE, coords.shape[1]), dtype=coords.dtype)
    for iE in prange(nE):
        for iNE in prange(nNE):
            res[iE, iNE] = coords[topo[iE, iNE]]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def cell_coords(coords: ndarray, topo: ndarray) -> ndarray:
    """
    Returns coordinates of a single cell from a coordinate
    array and a topology array.

    Parameters
    ----------
    coords: numpy.ndarray
        2d array of shape (nP, nD) of vertex coordinates.
        nP : number of points
        nD : number of dimensions of the model space
    topo: numpy.ndarray
        1d array of vertex indices of shape (nNE,).
        nNE : number of nodes per element

    Returns
    -------
    numpy.ndarray
        2d NumPy array of (nNE, nD) of coordinates for all nodes of all cells
        according to the argument 'topo'.

    Notes
    -----
    The array 'coords' must be fully populated up to the maximum index
    in 'topo'. (len(coords) >= (topo.max() + 1))
    """
    nNE = len(topo)
    res = np.zeros((nNE, coords.shape[1]), dtype=coords.dtype)
    for iNE in prange(nNE):
        res[iNE] = coords[topo[iNE]]
    return res


@njit(nogil=True, cache=__cache)
def cell_center_2d(ecoords: ndarray) -> ndarray:
    """
    Returns the center of a 2d cell.

    Parameters
    ----------
    ecoords: numpy.ndarray
        2d coordinate array of the element. The array has as many rows,
        as the number of nodes of the cell, and two columns.

    Returns
    -------
    numpy.ndarray
        1d coordinate array.
    """
    return np.array(
        [np.mean(ecoords[:, 0]), np.mean(ecoords[:, 1])], dtype=ecoords.dtype
    )


@njit(nogil=True, cache=__cache)
def cell_center(coords: ndarray) -> ndarray:
    """
    Returns the center of a single cell.

    Parameters
    ----------
    ecoords: numpy.ndarray
        2d coordinate array of the element. The array has as many rows,
        as the number of nodes of the cell, and three columns.

    Returns
    -------
    numpy.ndarray
        1d coordinate array.
    """
    return np.array(
        [np.mean(coords[:, i]) for i in range(coords.shape[1])],
        dtype=coords.dtype,
    )


def cell_centers_bulk(coords: ndarray, topo: ndarray) -> ndarray:
    """
    Returns coordinates of the centers of cells of the same kind.

    Parameters
    ----------
    coords: numpy.ndarray
        2d coordinate array.
    topo: numpy.ndarray
        2d point-based topology array.

    Returns
    -------
    numpy.ndarray
        2d coordinate array.
    """
    return np.mean(cells_coords(coords, topo), axis=1)


def cell_centers_bulk2(ecoords: ndarray) -> ndarray:
    """
    Returns coordinates of the centers of cells of the same kind.

    Parameters
    ----------
    ecoords: numpy.ndarray
        3d coordinate array of element coordinates.

    Returns
    -------
    numpy.ndarray
        2d coordinate array.
    """
    return np.mean(ecoords, axis=1)


@njit(nogil=True, parallel=True, cache=__cache)
def _nodal_distribution_factors_csr_(topo: csr_matrix, w: ndarray) -> ndarray:
    """
    The j-th factor of the i-th row is the contribution of
    element i to the j-th node. Assumes zeroed and tight indexing.

    Parameters
    ----------
    topo: :class:`~sigmaepsilon.math.linalg.sparse.csr.csr_matrix`
        2d integer topology array as a CSR matrix.
    w: numpy.ndarray
        The weights of the cells.
    """
    nE = topo.shape[0]
    indptr = topo.indptr
    data = topo.data.astype(np.int32)
    factors = np.zeros(len(data), dtype=w.dtype)
    nodal_w = np.zeros(data.max() + 1, dtype=w.dtype)
    for iE in range(nE):
        nodal_w[data[indptr[iE] : indptr[iE + 1]]] += w[iE]
    for iE in prange(nE):
        _i = indptr[iE]
        i_ = indptr[iE + 1]
        n = i_ - _i
        for j in prange(n):
            i = _i + j
            factors[i] = w[iE] / nodal_w[data[i]]
    return factors


@njit(nogil=True, parallel=True, cache=__cache)
def _nodal_distribution_factors_dense_(topo: ndarray, w: ndarray) -> ndarray:
    """
    The j-th factor of the i-th row is the contribution of
    element i to the j-th node. Assumes zeroed and tight indexing.

    Parameters
    ----------
    topo: numpy.ndarray
        2d integer topology array.
    w: numpy.ndarray
        The weights of the cells.
    """
    factors = np.zeros(topo.shape, dtype=w.dtype)
    nodal_w = np.zeros(topo.max() + 1, dtype=w.dtype)
    for iE in range(topo.shape[0]):
        nodal_w[topo[iE]] += w[iE]
    for iE in prange(topo.shape[0]):
        for jNE in prange(topo.shape[1]):
            factors[iE, jNE] = w[iE] / nodal_w[topo[iE, jNE]]
    return factors


def nodal_distribution_factors(
    topo: Union[csr_matrix, ndarray], weights: ndarray
) -> Union[csr_matrix, ndarray]:
    """
    The j-th factor of the i-th row is the contribution of
    element i to the j-th node. Assumes zeroed and tight indexing.

    Parameters
    ----------
    topo: numpy.ndarray or :class:`~sigmaepsilon.math.linalg.sparse.csr.csr_matrix`
        2d integer topology array.
    w: numpy.ndarray
        The weights of the cells.

    Returns
    -------
    numpy.ndarray or :class:`~sigmaepsilon.math.linalg.sparse.csr.csr_matrix`
        A 2d matrix with a matching shape to 'topo'.

    See also
    --------
    :func:`~sigmaepsilon.mesh.data.polydata.PolyData.nodal_distribution_factors`
    """
    if isinstance(topo, ndarray):
        return _nodal_distribution_factors_dense_(topo, weights)
    elif isinstance(topo, csr_matrix):
        data = _nodal_distribution_factors_csr_(topo, weights)
        indptr = topo.indptr
        indices = topo.indices
        shape = topo.shape
        return csr_matrix(data, indices, indptr, shape)
    else:
        t = type(topo)
        raise TypeError(f"Type {t} is not recognized as a topology.")


@njit(nogil=True, parallel=True, cache=__cache)
def distribute_nodal_data_bulk(data: ndarray, topo: ndarray, ndf: ndarray) -> ndarray:
    """
    Distributes nodal data to the cells for the case when the topology
    of the mesh is dense. The parameter 'ndf' controls the behaviour
    of the distribution.

    Parameters
    ----------
    data: numpy.ndarray
        2d array of shape (nP, nX), the data defined on points.
    topo: numpy.ndarray
        2d integer array of shape (nE, nNE), describing the topology.
    ndf: numpy.ndarray
        2d float array of shape (nE, nNE), describing the distribution
        of cells to the nodes.

    Returns
    -------
    numpy.ndarray
        A 3d float array of shape (nE, nNE, nX).
    """
    nE, nNE = topo.shape
    res = np.zeros((nE, nNE, data.shape[1]))
    for iE in prange(nE):
        for jNE in prange(nNE):
            res[iE, jNE] = data[topo[iE, jNE]] * ndf[iE, jNE]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def distribute_nodal_data_sparse(
    data: ndarray, topo: ndarray, cids: ndarray, ndf: csr_matrix
) -> ndarray:
    """
    Distributes nodal data to the cells for the case when the topology of the
    mesh is sparse. The parameter 'ndf' controls the behaviour of the distribution.

    Parameters
    ----------
    data: numpy.ndarray
        2d array of shape (nP, nX), the data defined on points.
    topo: numpy.ndarray
        2d integer array of shape (nE, nNE), describing the topology.
    cids: numpy.ndarray
        A 1d integer array describing the indices of the cells.
    ndf: csr_matrix
        2d float array of shape (nE, nNE), describing the distribution
        of cells to all nodes in the mesh.

    Returns
    -------
    numpy.ndarray
        A 3d float array of shape (nE, nNE, nX).
    """
    nE, nNE = topo.shape
    indptr = ndf.indptr
    ndfdata = ndf.data
    res = np.zeros((nE, nNE, data.shape[1]))
    for iE in prange(nE):
        ndf_e = ndfdata[indptr[cids[iE]] : indptr[cids[iE] + 1]]
        for jNE in prange(nNE):
            res[iE, jNE] = data[topo[iE, jNE]] * ndf_e[jNE]
    return res


@njit(nogil=True, parallel=True, fastmath=True, cache=__cache)
def collect_nodal_data(
    celldata: ndarray, topo: ndarray, cids: ndarray, ndf: csr_matrix, res: ndarray
) -> ndarray:
    """
    Collects nodal data from data defined on nodes of cells.

    Parameters
    ----------
    celldata: numpy.ndarray
        Data defined on nodes of cells. It can be any array with at
        least 2 dimensions with a shape (nE, nNE, ...), where nE and
        nNE are the number of cells and nodes per cell.
    topo: numpy.ndarray
        A 2d integer array describing the topology of several cells of
        the same kind.
    cids: numpy.ndarray
        A 1d integer array describing the indices of the cells.
    ndf: csr_matrix
        Nodal distribution factors for each node of each cell in 'topo'.
        This must contain values for all cells in a mesh, not just the
        ones for which cell data and topology is provided by 'celldata'
        and 'topo'.
    res: numpy.ndarray
        An array for the output. It must have a proper size, at lest up
        to the maximum node index in 'topo'.
    """
    nE, nNE = topo.shape
    indptr = ndf.indptr
    ndfdata = ndf.data
    for iE in range(nE):
        ndf_e = ndfdata[indptr[cids[iE]] : indptr[cids[iE] + 1]]
        for jNE in prange(nNE):
            res[topo[iE, jNE]] += celldata[iE, jNE] * ndf_e[jNE]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def explode_mesh_bulk(coords: ndarray, topo: ndarray) -> Tuple[ndarray]:
    """
    Turns an implicit representation of a mesh into an explicit one.

    .. note:
        This function is Numba-jittable in 'nopython' mode.

    Parameters
    ----------
    coords: numpy.ndarray
        A 2d coordinate array.
    topo: numpy.ndarray
        A 2d integer array describing the topology of several cells of
        the same kind.

    Returns
    -------
    numpy.ndarray
        A new coordinate array.
    numpy.ndarray
        A new topology array.
    """
    nE, nNE = topo.shape
    nD = coords.shape[1]
    coords_ = np.zeros((nE * nNE, nD), dtype=coords.dtype)
    topo_ = np.zeros_like(topo)
    for i in prange(nE):
        ii = i * nNE
        for j in prange(nNE):
            coords_[ii + j] = coords[topo[i, j]]
            topo_[i, j] = ii + j
    return coords_, topo_


@njit(nogil=True, parallel=True, cache=__cache)
def explode_mesh_data_bulk(
    coords: ndarray, topo: ndarray, data: ndarray
) -> Tuple[ndarray]:
    """
    Turns an implicit representation of a mesh into an explicit one
    and also data defined on the nodes of the cells to an 1d data
    array defined on the points of the new mesh.

    .. note:
        This function is Numba-jittable in 'nopython' mode.

    Parameters
    ----------
    coords: numpy.ndarray
        A 2d coordinate array.
    topo: numpy.ndarray
        A 2d integer array describing the topology of several cells of
        the same kind.
    data: numpy.ndarray
        A 2d array describing data on all nodes of the cells.

    Returns
    -------
    numpy.ndarray
        A new coordinate array.
    numpy.ndarray
        A new topology array.
    numpy.ndarray
        A new 1d data array.
    """
    nE, nNE = topo.shape
    nD = coords.shape[1]
    coords_ = np.zeros((nE * nNE, nD), dtype=coords.dtype)
    topo_ = np.zeros_like(topo)
    data_ = np.zeros(nE * nNE, dtype=coords.dtype)
    for i in prange(nE):
        ii = i * nNE
        for j in prange(nNE):
            coords_[ii + j] = coords[topo[i, j]]
            data_[ii + j] = data[i, j]
            topo_[i, j] = ii + j
    return coords_, topo_, data_


def explode_mesh(coords: ndarray, topo: ndarray, *, data=None):
    if data is None:
        return explode_mesh_bulk(coords, topo)
    elif isinstance(data, ndarray):
        return explode_mesh_data_bulk(coords, topo, data)
    else:
        raise NotImplementedError


@njit(nogil=True, parallel=True, cache=__cache)
def decompose(ecoords: ndarray, topo: ndarray, coords_out: ndarray) -> None:
    """
    Performes the inverse operation to coordinate explosion.
    Example usage at AxisVM domains. Works for all kinds of arrays.
    """
    for iE in prange(len(topo)):
        for jNE in prange(len(topo[iE])):
            coords_out[topo[iE][jNE]] = np.array(ecoords[iE][jNE])


@njit(nogil=True, parallel=True, cache=__cache)
def _avg_cell_data_1d_bulk_(data: np.ndarray, topo: np.ndarray):
    nE, nNE = topo.shape
    nD = data.shape[1]
    res = np.zeros((nE, nD), dtype=data.dtype)
    for iE in prange(nE):
        for jNE in prange(nNE):
            ind = topo[iE, jNE]
            for kD in prange(nD):
                res[iE, kD] += data[ind, kD]
        res[iE, :] /= nNE
    return res


def avg_cell_data(data: np.ndarray, topo: np.ndarray) -> ndarray:
    nR = len(data.shape)
    if nR == 2:
        res = _avg_cell_data_1d_bulk_(matrixform(data), topo)
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def jacobian_matrix_single(dshp: ndarray, ecoords: ndarray) -> ndarray:
    """
    Returns Jacobian matrix of local to global transformation for for one cell and
    multiple evaluation pointsd.

    Parameters
    ----------
    dshp: numpy.ndarray
        A 3d numpy array of shape (nP, nNE, nD), where nP, nNE and nD
        are the number of evaluation points, nodes and spatial dimensions.
    ecoords: numpy.ndarray
        A 2d numpy array of shape (nNE, nD), where nNE and nD
        are the number of nodes and spatial dimensions.

    Returns
    -------
    numpy.ndarray
        A 3d array of shape (nP, nD, nD).
    """
    nG, _, nD = dshp.shape
    jac = np.zeros((nG, nD, nD), dtype=dshp.dtype)
    for iG in prange(nG):
        jac[iG] = dshp[iG].T @ ecoords
    return jac


@njit(nogil=True, parallel=True, cache=__cache)
def jacobian_matrix_bulk(dshp: ndarray, ecoords: ndarray) -> ndarray:
    """
    Returns Jacobian matrices of local to global transformation
    for several cells and evaluation points.

    Parameters
    ----------
    dshp: numpy.ndarray
        A 3d numpy array of shape (nG, nNE, nD), where nG, nNE and nD
        are the number of integration points, nodes and spatial dimensions.
    ecoords: numpy.ndarray
        A 3d numpy array of shape (nE, nNE, nD), where nE, nNE and nD
        are the number of elements, nodes and spatial dimensions.

    Returns
    -------
    numpy.ndarray
        A 4d array of shape (nE, nP, nD, nD).
    """
    nE = ecoords.shape[0]
    nG, _, nD = dshp.shape
    jac = np.zeros((nE, nG, nD, nD), dtype=dshp.dtype)
    for iG in prange(nG):
        d = dshp[iG].T
        for iE in prange(nE):
            jac[iE, iG] = d @ ecoords[iE]
    return jac


@njit(nogil=True, parallel=True, cache=__cache)
def jacobian_matrix_bulk_1d(dshp: ndarray, ecoords: ndarray) -> ndarray:
    """
    Returns the Jacobian matrix for multiple cells (nE), evaluated at
    multiple (nP) points.

    Returns
    -------
    numpy.ndarray
        A 4d NumPy array of shape (nE, nP, 1, 1).

    Notes
    -----
    As long as the line is straight, it is a constant metric element,
    and 'dshp' is only required here to provide an output with a correct
    shape.
    """
    lengths = lengths_of_lines2(ecoords)
    nE = ecoords.shape[0]
    if len(dshp.shape) > 4:
        # variable metric element -> dshp (nE, nP, nNE, nDOF, ...)
        nP = dshp.shape[1]
    else:
        # constant metric element -> dshp (nP, nNE, nDOF, ...)
        nP = dshp.shape[0]
    res = np.zeros((nE, nP, 1, 1), dtype=dshp.dtype)
    for iE in prange(nE):
        res[iE, :, 0, 0] = lengths[iE] / 2
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def jacobian_det_bulk_1d(jac: ndarray) -> ndarray:
    """
    Calculates Jacobian determinants for 1d cells.

    Parameters
    ----------
    jac: numpy.ndarray
        4d float array of shape (nE, nG, 1, 1) for an nE number of
        elements and nG number of evaluation points.

    Returns
    -------
    numpy.ndarray
        A 2d array of shape (nE, nG) of jacobian determinants calculated
        for each element and evaluation points.
    """
    nE, nG = jac.shape[:2]
    res = np.zeros((nE, nG), dtype=jac.dtype)
    for iE in prange(nE):
        res[iE, :] = jac[iE, :, 0, 0]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def center_of_points(coords: ndarray) -> ndarray:
    """
    Returns the center of several points.

    Parameters
    ----------
    coords: numpy.ndarray
        A 2d coordinate array.
    """
    res = np.zeros(coords.shape[1], dtype=coords.dtype)
    for i in prange(res.shape[0]):
        res[i] = np.mean(coords[:, i])
    return res


@njit(nogil=True, cache=__cache)
def centralize(coords: ndarray) -> ndarray:
    """
    Centralizes coordinates of a point cloud.

    Parameters
    ----------
    coords: numpy.ndarray
        A 2d coordinate array.
    """
    nD = coords.shape[1]
    center = center_of_points(coords)
    coords[:, 0] -= center[0]
    coords[:, 1] -= center[1]
    if nD > 2:
        coords[:, 2] -= center[2]
    return coords


@njit(nogil=True, parallel=True, cache=__cache)
def lengths_of_lines(coords: ndarray, topo: ndarray) -> ndarray:
    """
    Returns lengths of several lines, where the geometry is
    defined implicitly.

    Parameters
    ----------
    coords: numpy.ndarray
        A 2d coordinate array.
    topo: numpy.ndarray
        A 2d topology array.
    """
    nE, nNE = topo.shape
    res = np.zeros(nE, dtype=coords.dtype)
    for i in prange(nE):
        res[i] = norm(coords[topo[i, nNE - 1]] - coords[topo[i, 0]])
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def lengths_of_lines2(ecoords: ndarray) -> ndarray:
    """
    Returns lengths of several lines, where line cooridnates
    are specified explicitly.

    Parameters
    ----------
    ecoords: numpy.ndarray
        A 3d numpy array of shape (nE, nNE, nD), where nE, nNE and nD
        are the number of elements, nodes and spatial dimensions.
    """
    nE, nNE = ecoords.shape[:2]
    res = np.zeros(nE, dtype=ecoords.dtype)
    _nNE = nNE - 1
    for i in prange(nE):
        res[i] = norm(ecoords[i, _nNE] - ecoords[i, 0])
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def distances_of_points(coords: ndarray) -> ndarray:
    """
    Calculates distances between a series of points.

    Parameters
    ----------
    coords: numpy.ndarray
        2d float array of shape (N, ...).

    Returns
    -------
    numpy.ndarray
        1d float array of shape (nP,).
    """
    nP = coords.shape[0]
    res = np.zeros(nP, dtype=coords.dtype)
    for i in prange(1, nP):
        res[i] = norm(coords[i] - coords[i - 1])
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def pcoords_to_coords(pcoords: ndarray, ecoords: ndarray, shp: ndarray) -> ndarray:
    nP = pcoords.shape[0]
    nE, nNE, nD = ecoords.shape
    res = np.zeros((nE, nP, nD), dtype=ecoords.dtype)
    for iE in prange(nE):
        for iP in prange(nP):
            for iNE in range(nNE):
                res[iE, iP, :] += ecoords[iE, iNE] * shp[iP, iNE]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def pcoords_to_coords_1d(pcoords: ndarray, ecoords: ndarray) -> ndarray:
    """
    Returns a flattened array of points, evaluated at multiple
    points and cells.

    Only for 1d cells.

    Parameters
    ----------
    pcoords: numpy.ndarray
        1d float array of length nP, coordinates in the range [0 , 1].
    ecoords: numpy.ndarray
        3d float array of shape (nE, 2+, nD) of cell coordinates.

    Notes
    -----
    It works for arbitrary topologies, but handles every cell as a line
    going from the firts to the last node of the cell.

    Returns
    -------
    numpy.ndarray
        2d float array of shape (nE * nP, nD).
    """
    nP = pcoords.shape[0]
    nE = ecoords.shape[0]
    nX = nE * nP
    res = np.zeros((nX, ecoords.shape[2]), dtype=ecoords.dtype)
    for iE in prange(nE):
        for jP in prange(nP):
            res[iE * nP + jP] = (
                ecoords[iE, 0] * (1 - pcoords[jP]) + ecoords[iE, -1] * pcoords[jP]
            )
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def norms(a: ndarray) -> ndarray:
    """
    Returns the Euclidean norms for the input data, calculated
    along axis 1.

    Parameters
    ----------
    a: numpy.ndarray
        2d array of data of shape (N, ...).

    Returns
    -------
    numpy.ndarray
        1d float array of shape (N, ).
    """
    nI = len(a)
    res = np.zeros(nI)
    for iI in prange(len(a)):
        res[iI] = np.dot(a[iI], a[iI])
    return np.sqrt(res)


@njit(nogil=True, parallel=True, cache=__cache)
def homogenize_nodal_values(data: ndarray, measure: ndarray) -> ndarray:
    """
    Calculates constant values for cells from existing data defined for
    each node, according to some measure.
    """
    nE, _, nDATA = data.shape  # nE, nNE, nDATA
    res = np.zeros((nE, nDATA), dtype=data.dtype)
    for i in prange(nE):
        for j in prange(nDATA):
            res[i, j] = np.sum(data[i, :, j]) / measure[i]
    return res


def is_cw(points: Iterable) -> bool:
    """
    Returns True, if the polygon described by the points is clockwise,
    False if it is counterclockwise.
    """
    area = 0
    n = len(points)
    for i in range(n):
        j = (i + 1) % n
        area += points[i][0] * points[j][1] - points[j][0] * points[i][1]
    return area < 0


def is_ccw(points: Iterable) -> bool:
    """
    Returns True, if the polygon described by the points is counterclockwise,
    False if it is clockwise.
    """
    return not is_cw(points)


@njit(nogil=True, parallel=True, cache=__cache)
def global_shape_function_derivatives(dshp: ndarray, jac: ndarray) -> ndarray:
    """
    Calculates derivatives of shape functions wrt. the local axes of cells
    using derivatives along master axes evaulated at some points in
    the interval [-1, 1], and jacobian matrices of local-to-global mappings.

    Parameters
    ----------
    dshp: numpy.ndarray
        Derivatives of an nNE number of shape functions evaluated at an nP number
        of points as a float array of shape (nP, nNE, nD).
    jac: numpy.ndarray
        Jacobian matrices, evaluated for a a nP number of points in nE number of cells
        and for nD spatial dimensions as a float array of shape (nE, nP, nD, nD).

    Returns
    -------
    numpy.ndarray
        4d float array of shape (nE, nP, nNE, nD).
    """
    nP, nNE, nD = dshp.shape
    nE = jac.shape[0]
    res = np.zeros((nE, nP, nNE, nD), dtype=dshp.dtype)
    for iE in prange(nE):
        for iP in prange(nP):
            invJ = np.linalg.inv(jac[iE, iP])
            res[iE, iP] = dshp[iP] @ invJ.T
    return res


def coords_to_3d(x: ndarray) -> ndarray:
    x = atleast2d(x, back=True)
    if (N := x.shape[-1]) == 3:
        return x
    else:
        res = np.zeros((x.shape[0], 3), dtype=x.dtype)
        if N == 2:
            res[:, :2] = x
        elif N == 1:
            res[:, 0] = x.flatten()
        return res



# -*- coding: utf-8 -*-
from ...config import __hasmatplotlib__

if not __hasmatplotlib__:  # pragma: no cover

    def parallel_mpl(*_, **__):
        raise ImportError(
            "You need Matplotlib for this. Install it with 'pip install matplotlib'. "
            "You may also need to restart your kernel and reload the package."
        )

    def aligned_parallel_mpl(*_, **__):
        raise ImportError(
            "You need Matplotlib for this. Install it with 'pip install matplotlib'. "
            "You may also need to restart your kernel and reload the package."
        )

else:
    from typing import Iterable, Hashable, Union, Optional

    import matplotlib as mpl
    import matplotlib.pyplot as plt
    from matplotlib.path import Path
    from matplotlib.patches import PathPatch
    import matplotlib.gridspec as gridspec
    from matplotlib.widgets import Slider
    from matplotlib.figure import Figure
    import numpy as np
    from numpy import ndarray

    from sigmaepsilon.deepdict import DeepDict
    from sigmaepsilon.core.formatting import float_to_str_sig as str_sig
    from sigmaepsilon.math import atleast1d

    def parallel_mpl(
        data: Union[dict, Iterable[ndarray], ndarray],
        *,
        labels: Optional[Union[Iterable[str], None]] = None,
        padding: Optional[float] = 0.05,
        colors: Optional[Union[Iterable[str], None]] = None,
        lw: Optional[float] = 0.2,
        bezier: Optional[bool] = True,
        figsize: Optional[Union[tuple, None]] = None,
        title: Optional[Union[str, None]] = None,
        ranges: Optional[Union[Iterable[float], None]] = None,
        return_figure: Optional[bool] = True,
        **_,
    ) -> Union[Figure, None]:
        """
        Parameters
        ----------
        data: Union[Iterable[numpy.ndarray], dict, numpy.ndarray]
            A list of numpy.ndarray for each column. Each array is 1d with a length of N,
            where N is the number of data records (the number of lines).
        labels: Iterable, Optional
            Labels for the columns. If provided, it must have the same length as `data`.
        padding: float, Optional
            Controls the padding around the axes.
        colors: list of float, Optional
            A value for each record. Default is None.
        lw: float, Optional
            Linewidth.
        bezier: bool, Optional
            If True, bezier curves are created instead of a linear polinomials.
            Default is True.
        figsize: tuple, Optional
            A tuple to control the size of the figure. Default is None.
        title: str, Optional
            The title of the figure.
        ranges: list of list, Optional
            Ranges of the axes. If not provided, it is inferred from
            the input values, but this may result in an error.
            Default is False.
        return_figure: bool, Optional
            If True, the figure is returned. Default is False.

        Example
        -------
        .. plot::
            :include-source: True

            from sigmaepsilon.mesh.plotting import parallel_mpl
            import numpy as np
            colors = np.random.rand(150, 3)
            labels = [str(i) for i in range(10)]
            values = [np.random.rand(150) for i in range(10)]
            parallel_mpl(
                values,
                labels=labels,
                padding=0.05,
                lw=0.2,
                colors=colors,
                title="Parallel Plot with Random Data",
            )
        """

        if isinstance(data, dict):
            if labels is None:
                labels = list(data.keys())
            ys = np.dstack(list(data.values()))[0]
        elif isinstance(data, np.ndarray):
            assert labels is not None
            ys = data.T
        elif isinstance(data, Iterable):
            assert labels is not None
            ys = np.dstack(data)[0]
        else:
            raise TypeError("Invalid data type!")

        ynames = labels
        N, nY = ys.shape

        figsize = (7.5, 3) if figsize is None else figsize
        fig, host = plt.subplots(figsize=figsize)

        if ranges is None:
            ymins = ys.min(axis=0)
            ymaxs = ys.max(axis=0)
            ranges = np.stack((ymins, ymaxs), axis=1)
        else:
            ranges = np.array(ranges)

        # make sure that upper and lower ranges are not equal
        for i in range(nY):
            rmin, rmax = ranges[i]
            if abs(rmin - rmax) < 1e-12:
                rmin -= 1.0
                rmax += 1.0
                ranges[i] = [rmin, rmax]
        ymins = ranges[:, 0]
        ymaxs = ranges[:, 1]

        dys = ymaxs - ymins
        ymins -= dys * padding
        ymaxs += dys * padding
        dys = ymaxs - ymins

        # transform all data to be compatible with the main axis
        zs = np.zeros_like(ys)
        zs[:, 0] = ys[:, 0]
        zs[:, 1:] = (ys[:, 1:] - ymins[1:]) / dys[1:] * dys[0] + ymins[0]

        axes = [host] + [host.twinx() for i in range(nY - 1)]
        for i, ax in enumerate(axes):
            ax.set_ylim(ymins[i], ymaxs[i])
            ax.spines["top"].set_visible(False)
            ax.spines["bottom"].set_visible(False)
            if ax != host:
                ax.spines["left"].set_visible(False)
                ax.yaxis.set_ticks_position("right")
                ax.spines["right"].set_position(("axes", i / (nY - 1)))

        host.set_xlim(0, nY - 1)
        host.set_xticks(range(nY))
        host.set_xticklabels(ynames, fontsize=8)
        host.tick_params(axis="x", which="major", pad=7)
        host.spines["right"].set_visible(False)
        host.xaxis.tick_top()

        if title is not None:
            host.set_title(title, fontsize=12)

        for j in range(N):
            if not bezier:
                # to just draw straight lines between the axes:
                host.plot(range(nY), zs[j, :], c=colors[j], lw=lw)
            else:
                # create bezier curves
                # for each axis, there will a control vertex at the point itself, one at 1/3rd towards the previous and one
                #   at one third towards the next axis; the first and last axis have one less control vertex
                # x-coordinate of the control vertices: at each integer (for the axes) and two inbetween
                # y-coordinate: repeat every point three times, except the first and last only twice
                verts = list(
                    zip(
                        [
                            x
                            for x in np.linspace(
                                0, len(ys) - 1, len(ys) * 3 - 2, endpoint=True
                            )
                        ],
                        np.repeat(zs[j, :], 3)[1:-1],
                    )
                )
                # for x,y in verts: host.plot(x, y, 'go') # to show the control points of the beziers
                codes = [Path.MOVETO] + [Path.CURVE4 for _ in range(len(verts) - 1)]
                path = Path(verts, codes)
                patch = PathPatch(path, facecolor="none", lw=lw, edgecolor=colors[j])
                host.add_patch(patch)

        if return_figure:
            return fig

    def aligned_parallel_mpl(
        data: Union[ndarray, dict],
        datapos: Iterable[float],
        *,
        yticks=None,
        labels=None,
        sharelimits=False,
        texlabels=None,
        xticksrotation=0,
        suptitle=None,
        slider=False,
        slider_label=None,
        hlines=None,
        vlines=None,
        y0=None,
        xoffset=0.0,
        yoffset=0.0,
        return_figure: Optional[bool] = True,
        wspace: Optional[float] = 0.4,
        **kwargs,
    ) -> Union[Figure, None]:
        """
        Parameters
        ----------
        data: numpy.ndarray or dict
            The values to plot. If it is a NumPy array, labels must be provided
            with the argument `labels`, if it is a sictionary, the keys of the
            dictionary are used as labels.
        datapos: Iterable[float]
            Positions of the provided data values.
        yticks: Iterable[float], Optional
            Positions of ticks on the vertical axes. Default is None.
        labels: Iterable, Optional
            An iterable of strings specifying labels for the datasets.
            Default is None.
        sharelimits: bool, Optional
            If True, the axes share limits of the vertical axes.
            Default is False.
        texlabels: Itrable[str], Optional
            TeX-formatted labels. If provided, it must have the same length as
            `labels`. Default is None.
        xticksrotation: int, Optional
            Rotation of the ticks along the vertical axes. Expected in degrees.
            Default is 0.
        suptitle: str, Optional
            See Matplotlib's docs for the details. Default is None.
        slider: bool, Optional
            If True, a slider is added to the figure for interactive plots.
            Default is False.
        slider_label: str, Optional
            A label for the slider. Only if `slider` is true. Default is None.
        hlines: Iterable[float], Optional
            A list of data values where horizontal lines are to be added to the axes.
            Default is None.
        vlines[float]: Iterable, Optional
            A list of data values where vertical lines are to be added to the axes.
            Default is None.
        y0: float or int, Optional
            Value for the vertical axis. Default is the average of the limits
            of the vertical axis (0.5*(datapos[0] + datapos[-1])).
        xoffset: float, Optional
            Margin of the plot in the vertical direction. Default is 0.
        yoffset: float, Optional
            Margin of the plot in the horizontal direction. Default is 0.
        wspace: float, Optional
            Spacing between the axes. Default is 0.4.
        **kwargs: dict, Optional
            Extra keyword arguments are forwarded to the creator of the matplotlib figure.
            Default is None.

        Example
        -------
        .. plot::
            :include-source: True

            from sigmaepsilon.mesh.plotting.mpl.parallel import aligned_parallel_mpl
            import numpy as np
            labels = ['a', 'b', 'c']
            values = np.array([np.random.rand(150) for _ in labels]).T
            datapos = np.linspace(-1, 1, 150)
            aligned_parallel_mpl(values, datapos, labels=labels, yticks=[-1, 1], y0=0.0)
        """
        # init
        fig = plt.figure(**kwargs)
        suptitle = "" if suptitle is None else suptitle
        fig.suptitle(suptitle)
        plotdata = DeepDict()
        axcolor = "lightgoldenrodyellow"
        ymin, ymax = np.min(datapos), np.max(datapos)

        hlines = [] if hlines is None else hlines
        vlines = [] if vlines is None else vlines

        # init slider
        if slider:
            if slider_label is None:
                slider_label = ""

            if y0 is None:
                y0 = 0.5 * (ymin + ymax)

        # read data
        if isinstance(data, dict):
            if labels is None:
                labels = list(data.keys())
        elif isinstance(data, np.ndarray):
            nData = data.shape[1]
            if labels is None:
                labels = list(map(str, range(nData)))
            data = {labels[i]: data[:, i] for i in range(nData)}

        for lbl in labels:
            plotdata[lbl]["values"] = data[lbl]

        # set min and max values
        _min, _max = [], []
        for lbl in labels:
            plotdata[lbl]["min"] = np.min(data[lbl])
            plotdata[lbl]["max"] = np.max(data[lbl])
            _min.append(plotdata[lbl]["min"])
            _max.append(plotdata[lbl]["max"])

        # set global min and max
        vmin = np.min(_min)
        vmax = np.max(_max)
        del _min
        del _max

        # setting up figure layout
        nData = len(labels)
        if slider:
            nAxes = nData + 1  # +1 for the Slider
        else:
            nAxes = nData

        width_ratios = [1 for i in range(nData)]

        if slider:
            width_ratios.append(0.15)

        spec = gridspec.GridSpec(
            ncols=nAxes,
            nrows=1,
            width_ratios=width_ratios,
            figure=fig,
            wspace=wspace,
            left=0.1,
        )

        # create axes
        for i in range(nData):
            plotid = int("{}{}{}".format(1, nAxes, i + 1))
            plotid = spec[0, i]
            ax = fig.add_subplot(plotid, facecolor=axcolor)
            ax.grid(False)
            ax.patch.set_edgecolor("black")
            ax.patch.set_linewidth(0.5)

            if i == 0:
                if yticks is not None:
                    ax.set_yticks(yticks)
                    ax.set_yticklabels([str_sig(val, sig=3) for val in yticks])
            else:
                ax.set_yticks([])
                ax.set_yticklabels([])

            if y0 is not None:
                hline = ax.axhline(y=y0, color="#d62728", linewidth=1)

            bbox = dict(boxstyle="round", ec="black", fc="yellow", alpha=0.8)
            txt = ax.text(
                0.0,
                0.0,
                "NaN",
                size=10,
                ha="center",
                va="center",
                visible=False,
                bbox=bbox,
            )

            # horizontal lines
            ax.axhline(y=yticks[0], color="black", linewidth=0.5, linestyle="-")
            ax.axhline(y=yticks[-1], color="black", linewidth=0.5, linestyle="-")
            for hl in hlines:
                ax.axhline(y=hl, color="black", linewidth=0.5, linestyle="-")

            # a vertical lines
            for vl in vlines:
                ax.axvline(x=vl, color="black", linewidth=0.5, linestyle="-")

            # store objects
            plotdata[labels[i]]["ax"] = ax
            plotdata[labels[i]]["text"] = txt
            if y0:
                plotdata[labels[i]]["hline"] = hline

        # create slider
        if slider:
            sliderax = fig.add_subplot(spec[0, nAxes - 1], fc=axcolor)
            slider_ = Slider(
                sliderax,
                slider_label,
                valmin=ymin,
                valmax=ymax,
                valinit=0.0,
                orientation="vertical",
                valfmt="%.3f",
                closedmin=True,
                closedmax=True,
            )

        def _approx_at_y(y: float, plotkey: Hashable):
            lines2D = plotdata[plotkey]["lines"]
            values = lines2D.get_xdata()
            locations = lines2D.get_ydata()
            return np.interp(y, locations, values)

        def _set_yval(y):
            for axkey in plotdata.keys():
                if "hline" in plotdata[axkey]:
                    plotdata[axkey]["hline"].set_ydata(atleast1d(y))

                v_at_y = _approx_at_y(y, axkey)
                txtparams = {
                    "visible": True,
                    "x": v_at_y,
                    "y": y,
                    "text": str_sig(v_at_y, sig=4),
                }
                plotdata[axkey]["text"].update(txtparams)
            fig.canvas.draw_idle()

        def _update_slider(y=None):
            if y is None:
                y = slider.val
            _set_yval(y)

        def _set_xlim(axs: mpl.axes, vmin: float, vmax: float):
            voffset = (vmax - vmin) * xoffset
            if abs(vmin - vmax) > 1e-7:
                axs.set_xlim(vmin - voffset, vmax + voffset)
            xticks = [vmin, vmax]
            axs.set_xticks(xticks)
            rotation = kwargs.get("rotation", xticksrotation)
            axs.set_xticklabels(
                [str_sig(val, sig=3) for val in xticks], rotation=rotation
            )

        def _set_ylim(axs: mpl.axes, vmin: float, vmax: float):
            voffset = (vmax - vmin) * yoffset
            axs.set_ylim(vmin - voffset, vmax + voffset)

        # plot axes
        for i, axkey in enumerate(plotdata.keys()):
            axis = plotdata[axkey]["ax"]

            # set limits
            if sharelimits == True:
                _set_xlim(axis, vmin, vmax)
            else:
                _set_xlim(axis, plotdata[axkey]["min"], plotdata[axkey]["max"])
            _set_ylim(axis, ymin, ymax)

            # set labels
            if texlabels is not None:
                axis.set_title(texlabels[i])
            else:
                axis.set_title(str(axkey))

            # plot
            lines = axis.plot(plotdata[axkey]["values"], datapos, picker=5)[0]
            plotdata[axkey]["lines"] = lines

        # connect events
        if slider:
            slider_.on_changed(_update_slider)
            fig._slider = (
                slider_  # to keep reference, otherwise slider is not responsive
            )

        if y0 is not None:
            _set_yval(y0)

        if return_figure:
            return fig


__all__ = ["parallel_mpl", "aligned_parallel_mpl"]



from typing import MutableMapping, Union, Dict, List, Tuple, Iterable, Any

import numpy as np
from numpy import ndarray
import awkward as ak
from awkward import Array as akarray
from scipy.sparse import csr_matrix as csr_scipy

from numba import njit, prange, types as nbtypes
from numba.typed import Dict as nbDict

from sigmaepsilon.math.linalg.sparse import csr_matrix, JaggedArray
from sigmaepsilon.math.arraysetops import unique2d
from sigmaepsilon.math import count_cols

from ...space import PointCloud
from ...topoarray import TopologyArray
from ...config import __hasnx__

if __hasnx__:
    import networkx as nx


__all__ = [
    "is_regular",
    "regularize",
    "count_cells_at_nodes",
    "cells_at_nodes",
    "nodal_adjacency",
    "unique_topo_data",
    "remap_topo",
    "detach_mesh_bulk",
    "rewire",
    "detach",
    "detach_mesh_data_bulk",
]


__cache = True
CoordsLike = Union[ndarray, PointCloud]
TopoLike = Union[ndarray, akarray, csr_matrix]
MappingLike = Union[ndarray, MutableMapping]
DoL = Dict[int, List[int]]

nbint32 = nbtypes.int32
nbint32A = nbint32[:]
nbint64 = nbtypes.int64
nbint64A = nbint64[:]


def rewire(topo: TopoLike, imap: MappingLike, invert: bool = False) -> Iterable:
    """
    Returns a new topology array. The argument 'imap' may be
    a dictionary or an array, that contains new indices for
    the indices in the old topology array.

    Parameters
    ----------
    topo: numpy.ndarray or sigmaepsilon.math.linalg.sparse.JaggedArray
        1d or 2d integer array representing topological data of a mesh.
    imap: MappingLike
        Inverse mapping on the index sets from global to local.
    invert: bool, Optional
        If `True` the argument `imap` describes a local to global
        mapping and an inversion takes place. In this case,
        `imap` must be a `numpy` array. Default is False.

    Returns
    -------
    TopoLike
        The same topology with the new numbering.
    """
    if invert:
        assert isinstance(imap, ndarray)
        imap = inds_to_invmap_as_dict(imap)
    if isinstance(topo, ndarray):
        if len(topo.shape) == 2:
            return remap_topo(topo, imap)
        elif len(topo.shape) == 1:
            return remap_topo_1d(topo, imap)
    elif isinstance(topo, TopologyArray):
        cuts, topo1d = topo.flatten(return_cuts=True)
        topo1d = remap_topo_1d(topo1d, imap)
        return TopologyArray(JaggedArray(topo1d, cuts=cuts))
    elif isinstance(topo, akarray):
        cuts = count_cols(topo)
        topo1d = ak.flatten(topo).to_numpy()
        topo1d = remap_topo_1d(topo1d, imap)
        return JaggedArray(topo1d, cuts=cuts)
    else:
        raise TypeError("Invalid topology with type <{}>".format(type(topo)))


@njit(nogil=True, parallel=True, cache=__cache)
def remap_topo(topo: ndarray, imap) -> ndarray:
    """
    Returns a new topology array. The argument 'imap' may be
    a dictionary or an array, that contains new indices for
    the indices in the old topology array.
    """
    nE, nNE = topo.shape
    res = np.zeros_like(topo)
    for iE in prange(nE):
        for jNE in prange(nNE):
            res[iE, jNE] = imap[topo[iE, jNE]]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def remap_topo_1d(topo1d: ndarray, imap) -> ndarray:
    """
    Returns a new topology array. The argument 'imap' may be
    a dictionary or an array, that contains new indices for
    the indices in the old topology array.
    """
    N = topo1d.shape[0]
    res = np.zeros_like(topo1d)
    for i in prange(N):
        res[i] = imap[topo1d[i]]
    return res


def is_regular(topo: TopoLike) -> bool:
    """
    Returns True if the topology is regular, in the meaning
    that the smallest node index is zero, and every integer
    is represented up to the maximum index.
    """
    if isinstance(topo, ndarray):
        return topo.min() == 0 and len(np.unique(topo)) == topo.max() + 1
    elif isinstance(topo, akarray):
        return np.min(topo) == 0 and len(unique2d(topo)) == np.max(topo) + 1
    elif isinstance(topo, csr_matrix):
        t = topo.data.astype(np.int32)
        return t.min() == 0 and len(np.unique(t)) == t.max() + 1
    elif isinstance(topo, csr_scipy):
        return np.min(t) == 0 and len(np.unique(t)) == np.max(t) + 1
    else:
        raise NotImplementedError


def regularize(topo: TopoLike) -> Tuple[TopoLike, ndarray]:
    """
    Returns a regularized topology and the unique indices.
    The returned topology array contains indices of the unique
    array.

    Parameters
    ----------
    topo : numpy.array or awkward.Array
        A topology array.

    Returns
    -------
    numpy.array or awkward.Array
        An array with a similar type as the input array.
    """
    if isinstance(topo, ndarray):
        unique, regular = np.unique(topo, return_inverse=True)
        regular = regular.reshape(topo.shape)
        return regular, unique
    elif isinstance(topo, akarray):
        unique, regular = unique2d(topo, return_inverse=True)
        return regular, unique
    elif isinstance(topo, csr_matrix):
        t = topo.data.astype(np.int32)
        unique, regular = np.unique(t, return_inverse=True)
        topo.data[:] = regular
        return topo, unique
    else:
        t = type(topo)
        raise NotImplementedError(f"Unknown type {t}")


@njit(nogil=True, parallel=False, fastmath=True, cache=__cache)
def _count_cells_at_nodes_reg_np_(topo: ndarray) -> ndarray:
    """
    Assumes a regular topology. Returns an array.
    """
    nE, nNE = topo.shape
    nN = topo.max() + 1
    count = np.zeros((nN), dtype=topo.dtype)
    for iE in prange(nE):
        for jNE in prange(nNE):
            count[topo[iE, jNE]] += 1
    return count


@njit(nogil=True, parallel=False, fastmath=True, cache=__cache)
def _count_cells_at_nodes_np_(topo: ndarray, nodeIDs: ndarray) -> Dict[int, int]:
    """
    Returns a dict{int : int} for the nodes in `nideIDs`.
    Assumes an irregular topology. The array `topo` must contain
    indices relative to `nodeIDs`. If the topology is regular,
    `nodeIDs == np.arange(topo.max() + 1)` is `True`.
    """
    nE, nNE = topo.shape
    count = dict()
    for i in range(len(nodeIDs)):
        count[nodeIDs[i]] = 0
    for iE in prange(nE):
        for jNE in prange(nNE):
            count[nodeIDs[topo[iE, jNE]]] += 1
    return count


@njit(nogil=True, parallel=False, fastmath=True, cache=__cache)
def _count_cells_at_nodes_reg_ak_(topo: akarray, nN: int) -> ndarray:
    """
    Assumes a regular topology. Returns an array.
    """
    ncols = count_cols(topo)
    nE = len(ncols)
    count = np.zeros((nN), dtype=np.int64)
    for iE in prange(nE):
        for jNE in prange(ncols[iE]):
            count[topo[iE][jNE]] += 1
    return count


@njit(nogil=True, parallel=False, fastmath=True, cache=__cache)
def _count_cells_at_nodes_ak_(topo: akarray, nodeIDs: ndarray) -> Dict[int, int]:
    """
    Returns a dict{int : int} for the nodes in `nideIDs`.
    Assumes an irregular topology. The array `topo` must contain
    indices relative to `nodeIDs`. If the topology is regular,
    `nodeIDs == np.arange(topo.max() + 1)` is `True`.
    """
    ncols = count_cols(topo)
    nE = len(ncols)
    count = dict()
    for i in range(len(nodeIDs)):
        count[nodeIDs[i]] = 0
    for iE in prange(nE):
        for jNE in prange(ncols[iE]):
            count[nodeIDs[topo[iE, jNE]]] += 1
    return count


@njit(nogil=True, fastmath=True, cache=__cache)
def _count_cells_at_nodes_reg_csr_(topo: csr_matrix) -> ndarray:
    """
    Assumes a regular topology. Returns an array.
    """
    indptr = topo.indptr
    data = topo.data.astype(np.int32)
    nE = indptr.shape[0] - 1
    nN = np.max(data) + 1
    count = np.zeros((nN), dtype=indptr.dtype)
    for iE in range(nE):
        _i = indptr[iE]
        i_ = indptr[iE + 1]
        n = i_ - _i
        for j in range(n):
            i = _i + j
            count[data[i]] += 1
    return count


@njit(nogil=True, fastmath=True, cache=__cache)
def _count_cells_at_nodes_csr_(topo: csr_matrix, nodeIDs: ndarray) -> Dict[int, int]:
    """
    Returns a dict{int : int} for the nodes in `nideIDs`.
    Assumes an irregular topology. The array `topo` must contain
    indices relative to `nodeIDs`. If the topology is regular,
    `nodeIDs == np.arange(topo.max() + 1)` is `True`.
    """
    indptr = topo.indptr
    data = topo.data.astype(np.int32)
    nE = indptr.shape[0] - 1
    count = dict()
    for i in range(len(nodeIDs)):
        count[nodeIDs[i]] = 0
    for iE in range(nE):
        _i = indptr[iE]
        i_ = indptr[iE + 1]
        n = i_ - _i
        for j in range(n):
            i = _i + j
            count[nodeIDs[data[i]]] += 1
    return count


def count_cells_at_nodes(topo: TopoLike, regular: bool = False) -> Union[ndarray, dict]:
    """
    Returns an array or a dictionary, that counts connecting
    elements at the nodes of a mesh.

    Parameters
    ----------
    topo: TopoLike
        2d numpy array describing the topoogy of a mesh.
    regular: bool, Optional
        A True value means that 'topo' has tight and zeroed indexing.
        In this case, the output is a NumPy array. If False, the output
        a dictionary.

    Returns
    -------
    count: numpy.ndarray or dict
        Number of connecting elements for each node in a mesh.
    """
    if not regular:
        if isinstance(topo, ndarray):
            topo, nodeIDs = regularize(topo)
            return _count_cells_at_nodes_np_(topo, nodeIDs)
        elif isinstance(topo, akarray):
            topo, nodeIDs = regularize(topo)
            return _count_cells_at_nodes_ak_(topo, nodeIDs)
        elif isinstance(topo, csr_matrix):
            topo, nodeIDs = regularize(topo)
            return _count_cells_at_nodes_csr_(topo, nodeIDs)
        else:
            raise NotImplementedError
    else:
        if isinstance(topo, ndarray):
            return _count_cells_at_nodes_reg_np_(topo)
        elif isinstance(topo, akarray):
            nN = np.max(topo) + 1
            return _count_cells_at_nodes_reg_ak_(topo, nN)
        elif isinstance(topo, csr_matrix):
            return _count_cells_at_nodes_reg_csr_(topo)
        else:
            raise NotImplementedError


@njit(nogil=True, cache=__cache)
def _cells_at_nodes_reg_np_(topo: ndarray):
    """Assumes a regular topology."""
    nE, nNE = topo.shape
    nN = topo.max() + 1
    count = _count_cells_at_nodes_reg_np_(topo)
    cmax = count.max()
    ereg = np.zeros((nN, cmax), dtype=topo.dtype)
    nreg = np.zeros((nN, cmax), dtype=topo.dtype)
    count[:] = 0
    for iE in range(nE):
        for jNE in range(nNE):
            ereg[topo[iE, jNE], count[topo[iE, jNE]]] = iE
            nreg[topo[iE, jNE], count[topo[iE, jNE]]] = jNE
            count[topo[iE, jNE]] += 1
    return count, ereg, nreg


@njit(nogil=True, cache=__cache)
def _cells_at_nodes_reg_ak_(topo: akarray, nN: int):
    """Assumes a regular topology."""
    ncols = count_cols(topo)
    nE = len(ncols)
    count = _count_cells_at_nodes_reg_ak_(topo, nN)
    cmax = count.max()
    ereg = np.zeros((nN, cmax), dtype=np.int64)
    nreg = np.zeros((nN, cmax), dtype=np.int64)
    count[:] = 0
    for iE in range(nE):
        for jNE in range(ncols[iE]):
            ereg[topo[iE][jNE], count[topo[iE][jNE]]] = iE
            nreg[topo[iE][jNE], count[topo[iE][jNE]]] = jNE
            count[topo[iE][jNE]] += 1
    return count, ereg, nreg


@njit(nogil=True, cache=__cache)
def _cells_at_nodes_reg_csr_(topo: csr_matrix):
    """Assumes a regular topology."""
    indptr = topo.indptr
    data = topo.data.astype(np.int64)
    nE = len(indptr) - 1
    count = _count_cells_at_nodes_reg_csr_(topo)
    nN = np.max(data) + 1
    cmax = count.max()
    ereg = np.zeros((nN, cmax), dtype=data.dtype)
    nreg = np.zeros((nN, cmax), dtype=data.dtype)
    count[:] = 0
    for iE in range(nE):
        _i = indptr[iE]
        i_ = indptr[iE + 1]
        n = i_ - _i
        for j in prange(n):
            i = _i + j
            ereg[data[i], count[data[i]]] = iE
            nreg[data[i], count[data[i]]] = j
            count[data[i]] += 1
    return count, ereg, nreg


@njit(nogil=True, cache=__cache)
def _nodal_cell_data_to_dicts_(
    count: ndarray, ereg: ndarray, nreg: ndarray, cellIDs: ndarray, nodeIDs: ndarray
) -> Tuple[Dict, Dict]:
    ereg_d = nbDict.empty(key_type=nbint64, value_type=nbint64A)
    nreg_d = nbDict.empty(key_type=nbint64, value_type=nbint64A)
    for i in range(len(count)):
        ereg_d[nodeIDs[i]] = cellIDs[ereg[i, : count[i]]]
        nreg_d[nodeIDs[i]] = nreg[i, : count[i]]
    return ereg_d, nreg_d


@njit(nogil=True, parallel=True, cache=__cache)
def _nodal_cell_data_to_spdata_(
    count: np.ndarray, ereg: np.ndarray, nreg: np.ndarray
) -> tuple:
    nE = ereg.max() + 1
    nN = len(count)
    N = np.sum(count)
    indices = np.zeros(N, dtype=count.dtype)
    data = np.zeros(N, dtype=count.dtype)
    indptr = np.zeros(nN + 1, dtype=count.dtype)
    indptr[1:] = np.cumsum(count)
    for i in prange(nN):
        indices[indptr[i] : indptr[i + 1]] = ereg[i, : count[i]]
        data[indptr[i] : indptr[i + 1]] = nreg[i, : count[i]]
    shape = (nN, nE)
    return data, indices, indptr, shape


def cells_at_nodes(
    topo: TopoLike,
    *_,
    frmt: str = None,
    assume_regular: bool = False,
    cellIDs: Iterable = None,
    return_counts: bool = False,
    **__,
) -> Any:
    """
    Returns data about element connectivity at the nodes of a mesh.

    Parameters
    ----------
    topo: numpy.ndarray or sigmaepsilon.math.linalg.sparse.JaggedArray
        A 2d array (either jagged or not) representing topological data of a mesh.
    frmt: str
        A string specifying the output format. Valid options are
        'jagged', 'csr', 'scipy-csr' and 'dicts'.
        See below for the details on the returned object.
    return_counts: bool
        Wether to return the numbers of connecting elements at the nodes
        as a numpy array. If format is 'raw', the
        counts are always returned irrelevant to this option.
    assume_regular: bool
        If the topology is regular, you can gain some speed with providing
        it as `True`. Default is `False`.
    cellIDs: numpy.ndarray
        Indices of the cells in `topo`. If nor `None`, format must be 'dicts'.
        Default is `None`.

    Returns
    -------
    If `return_counts` is `True`, the number of connecting elements for
    each node is returned as either a numpy array (if `cellIDs` is `None`)
    or a dictionary (if `cellIDs` is not `None`). If format is 'raw', the
    counts are always returned.

    `frmt` = None

        counts : np.ndarray(nN) - numbers of connecting elements, only returned
            if 'return_counts' is `True`

        ereg : np.ndarray(nN, nmax) - indices of connecting elements

        nreg : np.ndarray(nN, nmax) - node indices with respect to the
            connecting elements
    where

        nN - is the number of nodes,

        nmax - is the maximum number of elements that meet at a node, that is
            count.max()

    `frmt` = 'csr'

        counts(optionally) : np.ndarray(nN) - number of connecting elements, only returned
            if 'return_counts' is `True`

        csr : csr_matrix - sparse matrix in a numba-jittable csr format.
                            Column indices denote element indices, values
                            have the meaning of element node locations.

    `frmt` = 'scipy-csr' or 'csr-scipy'

        counts(optionally) : np.ndarray(nN) - number of connecting elements, only returned
            if 'return_counts' is `True`

        csr : csr_matrix - An instance of scipy.linalg.sparse.csr_matrix.
                            Column indices denote element indices, values
                            have the meaning of element node locations.
    `frmt` = 'dicts'

        counts(optionally) : np.ndarray(nN) - number of connecting elements, only returned
            if 'return_counts' is `True`

        ereg : numba Dict(int : int[:]) - indices of elements for each node
                                            index

        nreg : numba Dict(int : int[:]) - local indices of nodes in the
                                            connecting elements

    `frmt` = 'jagged'

        counts(optionally) : np.ndarray(nN) - number of connecting elements, only returned
            if 'return_counts' is `True`

        ereg : JaggedArray - indices of elements for each node index

        nreg : JaggedArray - local indices of nodes in the connecting elements

    """
    if cellIDs is not None:
        assert frmt == "dicts", (
            "If `cellIDs` is not None," + " output format must be 'dicts'."
        )

    if not assume_regular:
        if is_regular(topo):
            nodeIDs = None
        else:
            topo, nodeIDs = regularize(topo)
    else:
        nodeIDs = None

    if nodeIDs is not None:
        assert frmt == "dicts", "Only the format 'dicts' supports an irregular input!"

    if isinstance(topo, ndarray):
        counts, ereg, nreg = _cells_at_nodes_reg_np_(topo)
    elif isinstance(topo, akarray):
        nN = np.max(topo) + 1
        counts, ereg, nreg = _cells_at_nodes_reg_ak_(topo, nN)
    elif isinstance(topo, csr_matrix):
        counts, ereg, nreg = _cells_at_nodes_reg_csr_(topo)
    else:
        raise NotImplementedError

    frmt = "" if frmt is None else frmt

    if frmt in ["csr", "csr-scipy", "scipy-csr"]:
        data, indices, indptr, shape = _nodal_cell_data_to_spdata_(counts, ereg, nreg)
        csr = csr_matrix(data=data, indices=indices, indptr=indptr, shape=shape)
        if frmt == "csr":
            csr = csr_matrix(data=data, indices=indices, indptr=indptr, shape=shape)
        else:
            csr = csr_scipy((data, indices, indptr), shape=shape)
        if return_counts:
            return counts, csr
        return csr
    elif frmt == "dicts":
        if isinstance(topo, csr_matrix):
            if cellIDs is None:
                cellIDs = np.arange(len(topo.indptr) - 1).astype(int)
        cellIDs = np.arange(len(topo)).astype(int) if cellIDs is None else cellIDs
        nodeIDs = np.arange(len(counts)).astype(int) if nodeIDs is None else nodeIDs
        cellIDs = cellIDs.astype(np.int64)
        nodeIDs = nodeIDs.astype(np.int64)
        ereg, nreg = _nodal_cell_data_to_dicts_(counts, ereg, nreg, cellIDs, nodeIDs)
        if return_counts:
            return counts, ereg, nreg
        return ereg, nreg
    elif frmt == "jagged":
        data, indices, indptr, shape = _nodal_cell_data_to_spdata_(counts, ereg, nreg)
        ereg = JaggedArray(indices, cuts=counts)
        nreg = JaggedArray(data, cuts=counts)
        if return_counts:
            return counts, ereg, nreg
        return ereg, nreg

    return counts, ereg, nreg


@njit(nogil=True, cache=__cache)
def _nodal_adjacency_as_dol_np_(topo: ndarray, ereg: DoL) -> DoL:
    """Returns nodal adjacency as a dictionary of lists."""
    res = dict()
    for iP in ereg:
        res[iP] = np.unique(topo[ereg[iP], :])
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def _subtopo_1d_(
    topo1d: ndarray, cuts: ndarray, inds: ndarray, indptr: ndarray
) -> ndarray:
    nN = np.sum(cuts[inds])
    nE = len(inds)
    subindptr = np.zeros(nE + 1, dtype=cuts.dtype)
    subindptr[1:] = np.cumsum(cuts[inds])
    subtopo1d = np.zeros(nN, dtype=topo1d.dtype)
    for iE in prange(nE):
        subtopo1d[subindptr[iE] : subindptr[iE + 1]] = topo1d[
            indptr[inds[iE]] : indptr[inds[iE] + 1]
        ]
    return subtopo1d


@njit(nogil=True, cache=__cache)
def _nodal_adjacency_as_dol_ak_(topo1d: ndarray, cuts: ndarray, ereg: DoL) -> DoL:
    """Returns nodal adjacency as a dictionary of lists."""
    res = dict()
    nN = len(cuts)
    indptr = np.zeros(nN + 1, dtype=cuts.dtype)
    indptr[1:] = np.cumsum(cuts)
    for iP in ereg:
        res[iP] = np.unique(_subtopo_1d_(topo1d, cuts, ereg[iP], indptr))
    return res


@njit(nogil=True, parallel=False, fastmath=False, cache=__cache)
def dol_keys(dol: DoL) -> ndarray:
    nD = len(dol)
    res = np.zeros(nD, dtype=np.int64)
    c = 0
    for i in dol:
        res[c] = i
        c += 1
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def dol_to_jagged_data(dol: DoL) -> Tuple[ndarray, ndarray]:
    nD = len(dol)
    keys = dol_keys(dol)
    widths = np.zeros(nD, dtype=np.int64)
    for i in prange(nD):
        widths[i] = len(dol[keys[i]])
    indptr = np.zeros(nD + 1, dtype=np.int64)
    indptr[1:] = np.cumsum(widths)
    N = np.sum(widths)
    data1d = np.zeros(N, dtype=np.int64)
    for i in prange(nD):
        data1d[indptr[i] : indptr[i + 1]] = dol[keys[i]]
    return widths, data1d


def detach(coords: CoordsLike, topo: TopoLike, inds: ndarray | None = None):
    """
    Given a topology array and the coordinate array it refers to,
    the function returns the coordinate array of the points involved
    in the topology, and a new topology array, with indices referencing
    the unique coordinate array.

    Parameters
    ----------
    coords: CoordsLike
        A 2d float array representing geometrical data of a mesh.
        If it is a `PointCloud` instance, indices may be included
        and parameter `inds` is obsolete.
    topo: TopoLike
        A 2d integer array (either jagged or not) representing topological
        data of a mesh.
    inds: ndarray, Optional
        Global indices of the coordinates, in `coords`. If provided, the
        coordinates of node `j` of cell `i` is accessible as

        ``coords[imap[topo[i, j]]``,

        where `imap` is a mapping from local to global indices, and gets
        automatically generated from `inds`.  Default is None.

    Returns
    -------
    ndarray
        NumPy array similar to `coords`, but possibly with less entries.
    TopoLike
        Integer array representing the topology, with a good cahnce of
        being jagged, depending on your input.
    """
    if isinstance(topo, ndarray):
        if inds is None and isinstance(coords, PointCloud):
            inds = coords.inds
        if isinstance(inds, ndarray):
            topo = rewire(topo, inds, True)
        return detach_mesh_bulk(coords, topo)
    elif hasattr(topo, "__array_function__"):
        if inds is None and isinstance(coords, PointCloud):
            inds = coords.inds
        if isinstance(inds, ndarray):
            topo = rewire(topo, inds, True)
        return detach_mesh_jagged(coords, topo)
    else:
        raise TypeError("Invalid topology with type <{}>".format(type(topo)))


def detach_mesh_jagged(coords: ndarray, topo: ndarray):
    """
    Given a topology array and the coordinate array it refers to,
    the function returns the coordinate array of the points involved
    in the topology, and a new topology array, with indices referencing
    the new coordinate array.
    """
    inds = np.unique(topo)
    cuts, topo1d = topo.flatten(return_cuts=True)
    imap = inds_to_invmap_as_dict(inds)
    topo1d = remap_topo_1d(topo1d, imap)
    topo = JaggedArray(topo1d, cuts=cuts)
    return coords[inds, :], topo


@njit(nogil=True, cache=__cache)
def detach_mesh_bulk(coords: ndarray, topo: ndarray):
    """
    Given a topology array and the coordinate array it refers to,
    the function returns the coordinate array of the points involved
    in the topology, and a new topology array, with indices referencing
    the new coordinate array.
    """
    inds = np.unique(topo)
    return coords[inds], remap_topo(topo, inds_to_invmap_as_dict(inds))


@njit(nogil=True, cache=__cache)
def detach_mesh_data_bulk(coords: ndarray, topo: ndarray, data: ndarray):
    """
    Given a subset of the topology of a mesh, the function returns the
    coordinates and nodal data of the points involved in the topology,
    and a new topology array, with indices referencing the new coordinate
    array.
    """
    inds = np.unique(topo)
    coords_ = coords[inds]
    data_ = data[inds]
    return coords_, data_, remap_topo(topo, inds_to_invmap_as_dict(inds))


@njit(nogil=True, cache=__cache)
def inds_to_invmap_as_dict(inds: np.ndarray):
    """
    Returns a mapping that maps global indices to local ones.

    Parameters
    ----------
    inds: numpy.ndarray
        An array of global indices.

    Returns
    -------
    dict
        Mapping from global to local.
    """
    res = dict()
    for i in range(len(inds)):
        res[inds[i]] = i
    return res


@njit(nogil=True, cache=__cache)
def arrays_to_imap_as_dict(source: np.ndarray, target: np.ndarray):
    """
    Turns to index array into a dicionary such that `result[source[i]] = target[i]`.

    Parameters
    ----------
    source: numpy.ndarray
        An index array.
    target: numpy.ndarray
        An index array.

    Returns
    -------
    dict
        Mapping from global to local.
    """
    res = dict()
    for i in range(len(source)):
        res[source[i]] = target[i]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def inds_to_invmap_as_array(inds: np.ndarray):
    """
    Returns a mapping that maps global indices to local ones
    as an array.

    Parameters
    ----------
    inds: numpy.ndarray
        An array of global indices.

    Returns
    -------
    numpy.ndarray
        Mapping from global to local.
    """
    res = np.zeros(inds.max() + 1, dtype=inds.dtype)
    for i in prange(len(inds)):
        res[inds[i]] = i
    return res


def nodal_adjacency(
    topo: TopoLike, *_, frmt: str = None, assume_regular: bool = False, **__
) -> Any:
    """
    Returns nodal adjacency information of a mesh.

    Parameters
    ----------
    topo: numpy.ndarray array or JaggedArray
        A 2d array (either jagged or not) representing topological data of a mesh.
    frmt: str
        A string specifying the output format. Valid options are
        'jagged', 'csr', 'nx' and 'scipy-csr'. See below for the details on the
        returned object.
    assume_regular: bool
        If the topology is regular, you can gain some speed with providing
        it as `True`. Default is `False`.

    Notes
    -----
    1) You need `networkx` to be installed for most of the functionality here.

    2) A node is adjacent to itself.

    Returns
    -------
    `frmt` = None
        A dictionary of numpy arrays for each node.
    `frmt` = 'csr'
        csr_matrix - A sparse matrix in a numba-jittable csr format.
    `frmt` = 'scipy-csr'
        An instance of scipy.linalg.sparse.csr_matrix.
    `frmt` = 'nx'
        A networkx Graph.
    `frmt` = 'jagged'
        A JaggedArray instance.
    """
    frmt = "" if frmt is None else frmt
    ereg, _ = cells_at_nodes(topo, frmt="dicts", assume_regular=assume_regular)
    if isinstance(topo, ndarray):
        dol = _nodal_adjacency_as_dol_np_(topo, ereg)
    elif isinstance(topo, akarray):
        cuts, topo1d = JaggedArray(topo).flatten(return_cuts=True)
        dol = _nodal_adjacency_as_dol_ak_(topo1d, cuts, ereg)
    if not __hasnx__ and frmt != "jagged":
        errorstr = "`networkx` must be installed for format '{}'."
        raise ImportError(errorstr.format(frmt))
    if frmt == "nx":
        return nx.from_dict_of_lists(dol)
    elif frmt == "scipy-csr":
        G = nx.from_dict_of_lists(dol)
        return nx.to_scipy_sparse_array(G).tocsr()
    elif frmt == "csr":
        G = nx.from_dict_of_lists(dol)
        csr = nx.to_scipy_sparse_array(G).tocsr()
        return csr_matrix(csr)
    elif frmt == "jagged":
        cuts, data1d = dol_to_jagged_data(dol)
        return JaggedArray(data1d, cuts=cuts)
    return dol


def unique_topo_data(topo3d: TopoLike) -> Tuple[ndarray, ndarray]:
    """
    Returns information about unique topological elements
    of a mesh. It can be used to return unique lines of a 2d
    mesh, unique faces of a 3d mesh, etc.

    Parameters
    ----------
    topo: numpy.ndarray
        Hierarchical topology array. The array must be 3 dimensional containing node
        indices for every node as a subarray. For instance for a 2d cell, the node
        indices of the j-th edge of the i-th element read as `topo[i, j]`. In general,
        the first axis runs for the elements, the second axis runs for edges (2d) or
        faces (3d).

    Returns
    -------
    numpy.ndarray
        The sorted unique topolical entities as integer arrays
        of node indices.
    numpy.ndarray
        Indices of the unique array, that can be used to
        reconstruct `topo`. See the examples.

    Examples
    --------
    Find unique edges of a mesh of Q4 quadrilaterals

    >>> from sigmaepsilon.mesh.grid import grid
    >>> from sigmaepsilon.mesh.utils.topodata import edges_Q4
    >>> import numpy as np

    >>> coords, topo = grid(size=(1, 1), shape=(10, 10), eshape='Q4')

    To get a 3d integer array listing all the edges of all quads:

    >>> edges3d = edges_Q4(topo)

    To find the unique edges of the mesh:

    >>> from sigmaepsilon.mesh.utils.topology import unique_topo_data
    >>> edges, edgeIDs = unique_topo_data(edges3d)

    Then, to reconstruct `edges3d`, do the following

    >>> edges3d_ = np.zeros_like(edges3d)
    >>> for i in range(edgeIDs.shape[0]):
    ...     for j in range(edgeIDs.shape[1]):
    ...         edges3d_[i, j, :] = edges[edgeIDs[i, j]]
    """
    if isinstance(topo3d, ndarray):
        nE, nD, nN = topo3d.shape
        topo3d = topo3d.reshape((nE * nD, nN))
        topo3d = np.sort(topo3d, axis=1)
        topo3d, topoIDs = np.unique(topo3d, axis=0, return_inverse=True)
        topoIDs = topoIDs.reshape((nE, nD))
        return topo3d, topoIDs
    elif isinstance(topo3d, JaggedArray):
        raise NotImplementedError



from numba import njit, prange
import numpy as np
from numpy import ndarray

__cache = True


@njit(nogil=True, cache=__cache)
def monoms_L2(r: float) -> ndarray:
    return np.array([1, r], dtype=float)


@njit(nogil=True, cache=__cache)
def shp_L2(r) -> ndarray:
    """
    Evaluates the shape functions at one location in the range [-1, 1].
    """
    return np.array([1 - r, 1 + r]) / 2


@njit(nogil=True, parallel=True, cache=__cache)
def shp_L2_multi(pcoords: ndarray) -> ndarray:
    nP = pcoords.shape[0]
    res = np.zeros((nP, 2), dtype=pcoords.dtype)
    for iP in prange(nP):
        res[iP, :] = shp_L2(pcoords[iP])
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def shape_function_matrix_L2_multi(pcoords: ndarray, ndof: int = 2) -> ndarray:
    nP = pcoords.shape[0]
    eye = np.eye(ndof, dtype=pcoords.dtype)
    res = np.zeros((nP, ndof, 2 * ndof), dtype=pcoords.dtype)
    for iP in prange(nP):
        shp = shp_L2(pcoords[iP])
        for iN in prange(2):
            res[iP, :, iN * ndof : (iN + 1) * ndof] = eye * shp[iN]
    return res


@njit(nogil=True, cache=__cache)
def dshp_L2(r) -> ndarray:
    return np.array([-1, 1]) / 2


@njit(nogil=True, parallel=True, cache=__cache)
def dshp_L2_multi(pcoords: ndarray) -> ndarray:
    nP = pcoords.shape[0]
    res = np.zeros((nP, 2, 1), dtype=pcoords.dtype)
    for iP in prange(nP):
        res[iP, :, 0] = dshp_L2(pcoords[iP])
    return res



from ..config import __hask3d__, __hasmatplotlib__
from ..helpers import exporters

if not (__hask3d__ and __hasmatplotlib__):  # pragma: no cover

    def to_k3d(*_, **__):
        raise ImportError(
            "You need both K3D and Matplotlib for this. Install it with 'pip install k3d matplotlib'. "
            "You may also need to restart your kernel and reload the package."
        )

else:
    from copy import copy
    from typing import Union, Iterable, Optional
    import warnings

    import k3d
    import numpy as np
    from numpy import ndarray
    import matplotlib as mpl

    from sigmaepsilon.math import minmax

    from ..data import PolyData, PolyCell, PointData
    from ..utils.topology import detach_mesh_data_bulk, detach_mesh_bulk

    def to_k3d(
        obj: PolyData,
        *,
        scene: Optional[Union[k3d.Plot, None]] = None,
        deep: Optional[bool] = True,
        config_key: Optional[Union[str, None]] = None,
        menu_visibility: Optional[bool] = True,
        cmap: Optional[Union[list, None]] = None,
        show_edges: Optional[bool] = True,
        scalars: Optional[Union[ndarray, None]] = None,
    ) -> k3d.Plot:
        """
        Returns the mesh as a k3d mesh object.

        :: warning:
            Calling this method raises a UserWarning inside the `traittypes`
            package saying "Given trait value dtype 'float32' does not match
            required type 'float32'." However, plotting seems to be fine.

        Returns
        -------
        object
            A K3D Plot Widget, which is a result of a call to `k3d.plot`.
        """
        if scene is None:
            scene = k3d.plot(menu_visibility=menu_visibility)

        source = obj.source()
        coords = source.coords()

        if isinstance(scalars, ndarray):
            color_range = minmax(scalars)
            color_range = [scalars.min() - 1, scalars.max() + 1]

        k3dparams = dict(wireframe=False)
        if config_key is None:
            config_key = obj.__class__._k3d_config_key_

        cellblocks: Iterable[PolyData[PointData, PolyCell]] = obj.cellblocks(
            inclusive=True, deep=deep
        )

        for b in cellblocks:
            NDIM = b.celltype.Geometry.number_of_spatial_dimensions
            params = copy(k3dparams)
            config = b._get_config_(config_key)
            params.update(config)

            if "color" in params:
                if isinstance(params["color"], str):
                    hexstr = mpl.colors.to_hex(params["color"])
                    params["color"] = int("0x" + hexstr[1:], 16)

            if cmap is not None:
                params["color_map"] = cmap

            if NDIM == 1:
                topo = b.cd.topology().to_numpy()

                if isinstance(scalars, ndarray):
                    c, d, t = detach_mesh_data_bulk(coords, topo, scalars)
                    params["attribute"] = d
                    params["color_range"] = color_range
                    params["indices_type"] = "segment"
                else:
                    c, t = detach_mesh_bulk(coords, topo)
                    params["indices_type"] = "segment"

                c = c.astype(np.float32)
                t = t.astype(np.uint32)
                scene += k3d.lines(c, t, **params)
            elif NDIM == 2:
                topo = b.cd.to_triangles()

                if isinstance(scalars, ndarray):
                    c, d, t = detach_mesh_data_bulk(coords, topo, scalars)
                    params["attribute"] = d
                    params["color_range"] = color_range
                else:
                    c, t = detach_mesh_bulk(coords, topo)

                c = c.astype(np.float32)
                t = t.astype(np.uint32)

                if "side" in params:
                    if params["side"].lower() == "both":  # pragma: no cover
                        warnings.warn(
                            "The parameter 'both' was a typo in the documentation of K3D "
                            "we made up for it on the cost of performance by adding surfaces twice, "
                            "but the correct way is to use 'double' instead."
                        )
                        params["side"] = "front"
                        scene += k3d.mesh(c, t, **params)
                        params["side"] = "back"
                        scene += k3d.mesh(c, t, **params)
                    else:
                        scene += k3d.mesh(c, t, **params)
                else:
                    scene += k3d.mesh(c, t, **params)

                if show_edges:
                    scene += k3d.mesh(c, t, wireframe=True, color=0)
            elif NDIM == 3:
                topo = b.surface().topology().to_numpy()

                if isinstance(scalars, ndarray):
                    c, d, t = detach_mesh_data_bulk(coords, topo, scalars)
                    params["attribute"] = d
                    params["color_range"] = color_range
                else:
                    c, t = detach_mesh_bulk(coords, topo)

                c = c.astype(np.float32)
                t = t.astype(np.uint32)
                scene += k3d.mesh(c, t, **params)

                if show_edges:
                    scene += k3d.mesh(c, t, wireframe=True, color=0)

        return scene


exporters["k3d"] = to_k3d

__all__ = ["to_k3d"]



from typing import (
    Union,
    Iterable,
    Protocol,
    runtime_checkable,
    TypeVar,
    Generic,
    Optional,
    Tuple,
    ClassVar,
)

from numpy import ndarray

from sigmaepsilon.math.linalg import ReferenceFrame as FrameLike
from sigmaepsilon.math.linalg.sparse import csr_matrix

from .geometry import GeometryProtocol
from ..space import CartesianFrame
from ..topoarray import TopologyArray


__all__ = [
    "PointDataProtocol",
    "CellDataProtocol",
    "PolyCellProtocol",
    "PolyDataProtocol",
    "PolyDataLike",
    "PointDataLike",
    "CellDataLike",
    "PolyCellLike",
]


PolyDataLike = TypeVar("PolyDataLike", bound="PolyDataProtocol")
PointDataLike = TypeVar("PointDataLike", bound="PointDataProtocol")
CellDataLike = TypeVar("CellDataLike", bound="CellDataProtocol")
PolyCellLike = TypeVar("PolyCellLike", bound="PolyCellProtocol")


@runtime_checkable
class PointDataProtocol(Protocol):  # pragma: no cover
    """
    A protocol class for storing point-related data in a mesh.
    """

    @property
    def id(self) -> ndarray:
        """Ought to return global ids of the points as an 1d integer array."""
        ...

    @property
    def frame(self) -> FrameLike:
        """Ought to return a frame of reference."""
        ...

    @property
    def x(self) -> ndarray:
        """Ought to return the coordinates of the associated pointcloud
        as a 2d float array, where the first axis runs along the points."""
        ...

    def pull(self) -> ndarray:
        """Collects data at the points from the cells meeting at thenodes
        and aggregates it"""
        ...


@runtime_checkable
class CellDataProtocol(Generic[PolyDataLike, PointDataLike], Protocol):
    """
    A generic protocol class for storing cell-related data in a mesh.
    """

    @property
    def id(self) -> ndarray:
        """Ought to return global ids of the cells."""
        ...

    @property
    def frames(self) -> ndarray:
        """Ought to return the reference frames of the cells."""
        ...

    @property
    def nodes(self) -> ndarray:
        """
        Ought to return the topology of the cells as a 2d NumPy integer array.
        """
        ...

    @property
    def pointdata(self) -> PointDataLike:
        """Returns the hosting pointdata."""

    @property
    def container(self) -> PolyDataLike:
        """Returns the container object of the block."""

    @property
    def has_frames(self) -> bool:
        """
        Ought to return `True` if the cells are equipped with frames,
        `False` if they are not.
        """
        ...

    @property
    def has_nodes(self) -> bool:
        """
        Ought to return `True` if the cells are equipped with nodes,
        `False` if they are not.
        """
        ...


class PolyCellProtocol(
    CellDataProtocol[PolyDataLike, PointDataLike],
    Generic[PolyDataLike, PointDataLike],
    Protocol,
):  # pragma: no cover
    """
    A generic protocol class for polygonal cell containers.
    """

    label: ClassVar[Optional[str]] = None
    Geometry: ClassVar[GeometryProtocol]

    @property
    def db(self) -> CellDataProtocol[PolyDataLike, PointDataLike]:
        """
        Returns the database of the block.
        """
        ...

    @db.setter
    def db(self, value: CellDataProtocol[PolyDataLike, PointDataLike]) -> None:
        """
        Sets the database of the block.
        """
        ...

    def local_coordinates(self) -> ndarray:
        """Ought to return the coordinates of the cells in their local
        coordinate systems."""
        ...

    def coords(self) -> ndarray:
        """Ought to return the coordiantes associated with the object."""
        ...

    def topology(self) -> TopologyArray:
        """Ought to return the topology associated with the object."""
        ...

    def measures(self) -> ndarray:
        """Ought to return meaninful measures for each cell."""
        ...

    def normals(self) -> ndarray:
        """Ought to return the normal vectors of the surface of the mesh."""
        ...

    def measure(self) -> float:
        """Ought to return a single measure for a collection of cells."""
        ...

    def to_simplices(self) -> Tuple[ndarray]:
        """Ought to return a triangular representation of the mesh."""
        raise NotImplementedError

    def jacobian_matrix(self) -> ndarray:
        """Ought to return meaninful measures for each cell."""
        ...

    def loc_to_glob(
        self, x: Union[Iterable, ndarray], ec: Optional[Union[ndarray, None]] = None
    ) -> ndarray:
        """
        Maps local coordinates in the master domain to global cooridnates.
        The basis of the transformation is 'ec', which is the node coordinates
        array of the cells of shape (nE, nNE, nD), where 'nE', 'nNE' and 'nD' are
        the number of cells, nodes per cell and local spatial dimensions.
        """
        ...


@runtime_checkable
class PolyDataProtocol(
    Generic[PointDataLike, PolyCellLike], Protocol
):  # pragma: no cover
    """A generic protocol class for polygonal mesh containers."""

    @property
    def frame(self) -> Union[FrameLike, CartesianFrame]:
        """Ought to return the frame of the attached pointdata"""

    @property
    def pointdata(self) -> PointDataLike:
        """Ought to return the attached pointdata."""
        ...

    @property
    def celldata(self) -> PolyCellLike:
        """Ought to return the attached celldata."""
        ...

    def root(self: PolyDataLike, *args, **kwargs) -> PolyDataLike:
        """Ought to return the top level container."""
        ...

    def source(self: PolyDataLike, *args, **kwargs) -> Union[PolyDataLike, None]:
        """Ought to return the object that holds onto point data."""
        ...

    def coords(self, *args, **kwargs) -> ndarray:
        """Ought to return the coordiantes associated with the object."""
        ...

    def topology(self, *args, **kwargs) -> Union[ndarray, TopologyArray]:
        """Ought to return the topology associated with the object."""
        ...

    def nodal_distribution_factors(self, *args, **kwargs) -> Union[ndarray, csr_matrix]:
        """
        Ought to return nodal distribution factors for every node
        of every cell in the block.
        """
        ...

    def pointblocks(self: PolyDataLike, *args, **kwargs) -> Iterable[PolyDataLike]:
        """
        Ought to return PolyData blocks with attached PointData.
        """
        ...

    def cellblocks(self: PolyDataLike, *args, **kwargs) -> Iterable[PolyDataLike]:
        """
        Ought to return PolyData blocks with attached CellData.
        """
        ...
